# -*- coding: utf-8 -*-
"""Command-line interface functions for rpi2caster"""

# IMPORTS for text user interface
import io
import os
import readline
import glob

from . import exceptions as e
from . import constants as c

# Whether the debug mode is on (can be changed by setting module's attribute)
DEBUG_MODE = False
# Style modifiers for displaying bold, italic, smallcaps, inferior, superior
STYLE_MODIFIERS = {'roman': ' ',
                   'bold': '*',
                   'italic': '/',
                   'smallcaps': '#',
                   'subscript': '_',
                   'superscript': '^'}
# Some standard prompts
MSG_MENU = '[Enter] to go back to main menu...'
MSG_CONTINUE = '[Enter] to continue...'


def menu(options, header='', footer='', no_debug=False):
    """menu(options=[(name1, long1, func1), (name2, long2, func2)...],
                    header=foo,
                    footer=bar,
                    no_debug=False):

    A menu which takes four arguments:
    options list - contains tuples for each option:
        (description, long_description, function)
    header - string to be displayed above,
    footer - string to be displayed below,
    no_debug - True if we don't want to tell the user
        that the program is in debugging mode.

    After choice is made, return the command.

    Set up vars for conditional statements,
    and lists for appending new items.

    choices - options to be entered by user
    """
    # Clear the screen, display header and add two empty lines
    clear()
    if header:
        print(header, end='\n\n')
    # Get the first option - this will be listed last
    try:
        (zero_function, zero_desc, zero_long_desc) = options[0]
        functions = [zero_function]
    except IndexError:
        raise e.ExitProgram
    # Display all the options
    # Tab indent, option number, option name (not processing option 0 yet!)
    for i, (function, desc, long_desc) in enumerate(options):
        if i > 0:
            functions.append(function)
            print('\t %i : %s \n\t\t %s \n' % (i, desc, long_desc))
    # Option 0 is displayed last, add some whitespace around it
    print('\n\t %i : %s \n\t\t %s \n' % (0, zero_desc, zero_long_desc))
    # Print footer, if defined
    if footer:
        print(footer, end='\n\n')
    if DEBUG_MODE and not no_debug:
        print('The program is now in debugging mode!', end='\n\n')
    # Add an empty line to separate prompt
    print('\n')
    # Ask for user input
    choice_number = -1
    # Get only available options and exclude non-numeric strings
    while choice_number not in range(len(options)):
        # Wait until user enters proper data
        your_choice = input('Your choice: ')
        try:
            choice_number = int(your_choice)
        except ValueError:
            # Entered anything non-digit - repeat
            your_choice = ''
    # At last, we have chosen a valid option...
    # Return a corresponding value - which is option
    return functions[choice_number]


def clear():
    """Clears the screen"""
    os.system('clear')


def display(*args, **kwargs):
    """Displays info for the user - print all in one line"""
    print(*args, **kwargs)


def display_header(text, symbol='-'):
    """Displays a header banner"""
    dash_line = symbol * len(text)
    print('\n\n' + dash_line + '\n' + text + '\n' + dash_line + '\n')


def display_parameters(data):
    """Displays the parameters by section (given as a dictionary):
    {header1: [(param1_val1, param1_desc1), (param1_val2, param1_desc2)...],
     header2: [(param2_val1, param2_desc1), (param2_val2, param2_desc2)...]...}
     a section will be displayed if there are parameters to display;
     a parameter will be displayed if its value evaluates to True."""
    # {header: [(par1, desc1), (par2, desc2)], header2: [(...)]}
    for key in data:
        parameters = '\n'.join(['%s: %s' % (desc, value)
                                for (value, desc) in data[key] if value])
        if parameters:
            display_header(key)
            print(parameters)


def debug_info(*args, **kwargs):
    """Prints debug messages to screen if in debug mode"""
    if DEBUG_MODE:
        print('DEBUG: ', *args, **kwargs)


def debug_enter_data(prompt):
    """For debug-specific data"""
    if DEBUG_MODE:
        return input('DEBUG: ' + prompt)


def debug_pause(msg1='', msg2=MSG_CONTINUE):
    """For debug confirmations"""
    if DEBUG_MODE:
        input('DEBUG: ' + msg1 + ' - ' + msg2)


def pause(msg1='', msg2=MSG_CONTINUE):
    """Waits until user presses return"""
    input(msg1 + '\n' + msg2)


def enter_data(prompt, datatype=str):
    """Enter a value and convert it to the specific datatype"""
    value = ''
    while not value:
        value = input(prompt)
        try:
            value = datatype(value)
        except ValueError:
            print('Incorrect value or data type!')
            value = ''
    return value


def enter_data_or_blank(prompt, datatype=str):
    """enter_data_or_blank:

    Enter a value or leave blank, try to convert to the specified datatype
    """
    while True:
        value = input(prompt)
        if not value:
            return value
        try:
            return datatype(value)
        except ValueError:
            print('Incorrect value or data type!')


def _format_display(character, style):
    """format_display:

    Uses ANSI escape sequences to alter the appearance of the character
    displayed in the matrix case layout.

    Temporarily unused - until bugfix for 24574.
    """
    style_codes = {'roman': 0,
                   'bold': 1,
                   'italic': 33,
                   'smallcaps': 34,
                   'subscript': 35,
                   'superscript': 36}
    closing_sequence = '\033[0m'
    starting_sequence = '\033[' + str(style_codes.get(style, 0)) + 'm'
    character = starting_sequence + character + closing_sequence
    print(character)
    return character


def format_display(character, style):
    """This is a placeholder to be used until Python bug 24574 is fixed"""
    return STYLE_MODIFIERS.get(style, '') + character


def choose_one_style():
    """Chooses one style from available ones"""
    print('Choose a text style. Leave blank for roman.'
          'Available options: [r]oman, [b]old, [i]talic, [s]mall caps,\n'
          '[l]ower index (a.k.a. subscript, inferior), '
          '[u]pper index (a.k.a. superscript, superior).')
    style = enter_data_or_blank('Style?: ', str) or 'r'
    return c.STYLES.get(style, 'roman')


def choose_styles():
    """Chooses one or more styles and returns a list of them"""
    print('Choose one or more text styles, e.g. roman and small caps.\n'
          'Available options: [r]oman, [b]old, [i]talic, [s]mall caps,\n'
          '[l]ower index (a.k.a. subscript, inferior), '
          '[u]pper index (a.k.a. superscript, superior).\n'
          'Leave blank for roman only.')
    styles_string = enter_data_or_blank('Styles?: ', str) or 'r'
    return list({c.STYLES[char] for char in styles_string if char in c.STYLES})


def tab_complete(text, state):
    """tab_complete(text, state):

    This function enables tab key auto-completion when you
    enter the filename.
    """
    return (glob.glob(text+'*')+[None])[state]


def enter_input_filename():
    """Allows to enter the input filename and checks if it is readable"""
    # Set readline parameters
    readline.set_completer_delims(' \t\n;')
    readline.parse_and_bind('tab: complete')
    readline.set_completer(tab_complete)
    # Enter the input filename; check if the file is readable
    while True:
        prompt = '\nEnter the input file name (leave blank to abort): '
        filename = enter_data_or_blank(prompt)
        if not filename:
            return False
        filename = os.path.realpath(filename)
        try:
            with io.open(filename, 'r'):
                return filename
        except (IOError, FileNotFoundError):
            print('Wrong filename or file not readable!\n')


def enter_output_filename():
    """Allows user to enter output filename (without checking if readable)"""
    # Set readline parameters
    readline.set_completer_delims(' \t\n;')
    readline.parse_and_bind('tab: complete')
    readline.set_completer(tab_complete)
    # Enter the output filename; no check here
    prompt = '\nEnter the input file name (leave blank to abort): '
    filename = enter_data_or_blank(prompt)
    if filename:
        return os.path.realpath(filename)
    else:
        return False


def simple_menu(message, options):
    """Simple menu:

    A simple menu where user is asked what to do.
    Wrong choice points back to the menu.

    Message: prompt - string displayed on screen;
    options: a dict {ans1:opt1, ans2:opt2...}.
    """
    ans = ''
    while True:
        ans = input(message)
        if ans in options:
            return options[ans]
        elif ans.lower() in options:
            return options[ans.lower()]
        elif ans.upper() in options:
            return options[ans.upper()]
        else:
            pass


def confirm(question, default=None):
    """Asks a simple question with yes or no answers.
    Returns True for yes and False for no."""
    options = {'Y': True, 'N': False}
    def_string = ''
    if default is True or default is False:
        options[''] = default
        def_string = ', default: %s' % (default and 'Y' or 'N')
    return simple_menu('%s [Y / N%s]: ' % (question, def_string), options)


def display_diecase_layout(diecase):
    """Shows a layout for a given diecase ID, unit values for its
    assigned wedge, or the typical S5 if not specified."""
    def displayed_char(matrix):
        """Modifies matrix char for displaying"""
        spaces_symbols = {'_': '▣', ' ': '□', '': ' '}
        formatted_char = matrix.char
        for style in matrix.styles:
            formatted_char = format_display(formatted_char, style)
        return spaces_symbols.get(matrix.char, formatted_char)

    def row_units(row):
        """Gets wedge unit value for the row"""
        return str(diecase.wedge.units[row])

    matrices = [mat for mat in diecase]
    cols_set = {matrix.column for matrix in matrices}
    rows_set = {matrix.row for matrix in matrices}
    col_numbers = ((16 in rows_set or 'NI' in cols_set or 'NL' in cols_set) and
                   c.COLUMNS_17 or c.COLUMNS_15)
    # If row 16 found - generate 16 rows; else 15
    row_numbers = [x for x in range(1, 16 in rows_set and 17 or 16)]
    # Generate a header with column numbers
    header = ('|Row|' + ''.join([col.center(4) for col in col_numbers]) +
              '|' + 'Units'.center(7) + '|')
    # "-----" line in the table
    separator = '—' * len(header)
    # A row with only spaces and vertical lines in it
    empty_row = ('|' + ' ' * 3 + '|' + ' ' * 4 * len(col_numbers) + '|' +
                 ' ' * 7 + '|')
    # Initialize the displayed layout
    table = [separator, header, separator, empty_row]
    # Process each row
    for row_num in row_numbers:
        # Start with row number...
        row = ['|' + str(row_num).center(3) + '|']
        # Add only characters and styles, center chars to 4
        row.extend([displayed_char(mat).center(4)
                    for column_num in col_numbers for mat in matrices
                    if mat.column == column_num and mat.row == row_num])
        table.append(''.join(row) + '|' + row_units(row_num).center(7) + '|')
        table.append(empty_row)
    # Add the header at the bottom
    table.extend([separator, header, separator])
    # We can display it now
    print('\nDiecase ID: %s' % diecase)
    print('Stopbar / wedge: %s\n' % diecase.wedge)
    for row in table:
        print(row)
    # Explanation of symbols
    print('\nExplanation:', '□ = low space, ▣ = high space',
          '*a = bold, /a = italic, #a = small caps',
          '_a = subscript (inferior), ^a = superscript (superior)',
          sep='\n', end='\n\n')


def edit_matrix(matrix):
    """Edits the matrix data"""
    display_parameters({'Matrix details': matrix.parameters})
    prompt = ('Character (" "=low space, "_"=high space, '
              'blank to leave unchanged, ctrl-C to exit) ? : ')
    matrix.char = enter_data_or_blank(prompt) or matrix.char
    if not matrix.char:
        return matrix
    matrix.styles = choose_styles()
    units = matrix.units or matrix.row_units()
    prompt = 'Unit width? (current: %s) : ' % units
    matrix.units = enter_data_or_blank(prompt, int) or units
    return matrix


def edit_diecase_layout(diecase):
    """Edits a matrix case layout, row by row, matrix by matrix.
    Allows to enter a position to be edited. """
    def get_matrix(coordinates):
        """Gets matrix data for given coordinates."""
        return [mat for mat in diecase if mat.code == coordinates.upper()][0]

    def all_rows_mode():
        """Row-by-row editing - all cells in row 1, then 2 etc."""
        for mat in diecase:
            edit_matrix(mat)

    def all_columns_mode():
        """Column-by-column editing - all cells in column NI, NL, A...O"""
        # Rearrange the layout so that it's read column by column
        iter_mats = (mat for col in c.COLUMNS_17
                     for mat in diecase if mat.column == col)
        for mat in iter_mats:
            edit_matrix(mat)

    def single_row_mode(row):
        """Edits matrices found in a single row"""
        iter_mats = (mat for mat in diecase if mat.row == row)
        for mat in iter_mats:
            edit_matrix(mat)

    def single_column_mode(column):
        """Edits matrices found in a single column"""
        iter_mats = (mat for mat in diecase if mat.column == column)
        for mat in iter_mats:
            edit_matrix(mat)

    # Map unit values to rows
    # If the layout is empty, we need to initialize it
    prompt = ('Enter row number to edit all mats in a row,\n'
              'column number to edit all mats in a column,\n'
              'matrix coordinates to edit a single matrix,\n'
              'or choose edit mode: AR - all matrices row by row, '
              'AC - all matrices column by column.'
              '\nYour choice (or leave blank to exit) : ')
    while True:
        print('\nCurrent diecase layout:\n')
        display_diecase_layout(diecase)
        try:
            ans = input(prompt).upper()
            if ans == 'AR':
                all_rows_mode()
            elif ans == 'AC':
                all_columns_mode()
            elif ans in c.COLUMNS_17:
                single_column_mode(ans)
            elif ans in [str(x) for x in range(1, 17)]:
                single_row_mode(int(ans))
            elif ans:
                mat = get_matrix(ans)
                edit_matrix(mat)
            else:
                return diecase.matrices
        except (IndexError, KeyboardInterrupt, TypeError, AttributeError):
            pass


def enter_line_length():
    """Enter the line length, choose measurement units
    (for e.g. British or European measurement system). Return length in
    12-set units e.g. the standard Monotype measure."""
    prompt = ('Enter the desired line length and measurement unit:\nb'
              'c - cicero (.1776"), p - pica (.166"), dtp - DTP pica (.1667"),'
              '\n", in - inch, mm - millimeter, cm - centimeter?\n'
              '(default: c) : ')
    factor = 1.0
    # We need an ordered sequence here
    symbols = ['dtp', 'p', 'cm', 'c', 'mm', 'in', '"', '']
    units = {'dtp': 0.1667, 'p': 0.166, 'cm': 0.3937, 'mm': 0.03937,
             'c': 0.1776, '"': 1, 'in': 1, '': 0.1776}
    while True:
        raw_string = input(prompt).lower()
        try:
            for symbol in symbols:
                # Get the units
                if raw_string.endswith(symbol):
                    factor = units[symbol]
                    input_string = raw_string.replace(symbol, '')
                    input_string = input_string.strip()
                    break
            # Calculate length in inches and go on
            inches_length = float(input_string) * factor
            break
        except (TypeError, ValueError):
            print('Incorrect value - enter again...')
    # Return line length in 12-set units based on .1667" pica
    # Will be normalized in the typesetting routines
    return round(18 * 6 * inches_length)


def exit_program(*_):
    """Exit program:

    All objects call this method whenever they want to exit program.
    This is because we may do something specific in different UIs,
    so an abstraction layer may come in handy.
    """
    print('\n\nGoodbye!\n')
    exit()



# -*- coding: utf-8 -*-
"""
typesetting_functions:

Contains functions used for calculating line length, justification,
setting wedge positions, breaking the line etc.
"""
import io
from . import exceptions as e
from .global_settings import UI
from . import matrix_data
from . import wedge_data
from . import typesetting_data

COMMANDS = {'^00': 'roman', '^01': 'bold', '^02': 'italic',
            '^03': 'smallcaps', '^04': 'subscript', '^05': 'superscript',
            '^CR': 'align_left', '^CC': 'align_center', '^CL': 'align_right',
            '^CF': 'align_both'}


class TypesettingSession(object):
    """Glue layer for the whole typesetting session"""
    def __init__(self):
        # Instantiate all we need
        self.settings = Settings()
        self.translator = Translator()
        self.input_data = InputData()
        self.output_data = OutputData()
        # Make yourself a bridge between the objects
        self.settings.session = self
        self.translator.session = self
        self.input_data.session = self
        self.output_data.session = self

    def setup(self):
        """Sets up the parameters for the session"""
        pass

class Settings(object):
    """Typesetting job settings"""
    def __init__(self):
        self.wedge = wedge_data.Wedge()
        self.diecase = matrix_data.SelectDiecase()

    def set_ligatures(self):
        """Chooses the max ligature size i.e. how many characters can form
        a ligature"""
        prompt = 'Ligature: how many characters? '
        self.ligatures = abs(UI.enter_data(prompt, int))

    def session_setup(self, diecase_id):
        """Sets up initial typesetting session parameters:

        -diecase
        -line length and measurement units
        -default alignment
        -spaces
        -manual mode (more control) or automatic mode (less control, faster)
        """
        # Ask if the composing mode is manual or automatic
        self._manual_or_automatic()
        # Choose dominant stype
        self._choose_style()
        # Set it as the current style
        self.current_style = self.main_style

    def _manual_or_automatic(self):
        """Allows to choose if typesetting will be done with more user control.
        """
        # Manual control allows for tweaking more parameters during typesetting
        self.manual_control = UI.confirm('Use manual mode? (more control) ')
        if self.manual_control:
            # Choose unit shift: yes or no?
            self.unit_shift = UI.confirm('Do you use unit-shift? ')
            # Choose alignment mode
            self._choose_alignment()
            # Select the composition mode
            self.compose = self.manual_compose
            # Set custom spaces
            self._configure_spaces()

    def _choose_style(self):
        """Parses the diecase for available styles and lets user choose one."""
        available_styles = matrix_data.get_styles(self.diecase_layout)
        options = {str(i): style for i, style in
                   enumerate(available_styles, start=1)}
        # Nothing to choose from? Don't display the menu
        if len(options) == 1:
            self.main_style = options['1']
        # Otherwise, let user choose
        styles_list = ['%s : %s' % (x, options[x]) for x in sorted(options)]
        styles_list = ', '.join(styles_list)
        prompt = 'Choose a dominant style: %s ' % styles_list
        self.main_style = UI.simple_menu(prompt, options)

    @property
    def diecase(self):
        """Gets the diecase layout from the assigned diecase."""
        diecase = self.__dict__.get('_diecase', matrix_data.Diecase())
        return diecase.layout

    @diecase.setter
    def diecase(self, diecase):
        """Sets up the diecase and chooses the wedge."""
        self.__dict__['_diecase'] = diecase
        self.wedge = diecase.wedge


class InputData(object):
    """Gets the input text, parses it, generates a sequence of characters
    or commands"""
    def __init__(self):
        # Commands for activating the typesetting functions
        self.text = ''
        self.ligatures = 3

    def open_file(self, filename=''):
        """Opens a text file with text that will be typeset"""
        while True:
            # Choose file
            filename = filename or UI.enter_input_filename()
            if not filename:
                return False
            # Open it
            with io.open(filename, 'r') as text_file:
                self.text = '\n'.join(line for line in text_file)
                return True

    def parse_input(self, input_text):
        """Generates a sequence of characters from the input text.
        For each character, this function predicts what two next characters
        and one next character are."""
        # This variable will prevent yielding a number of subsequent chars
        # after a ligature or command has been found and yielded.
        skip_steps = 0
        spaces_names = {self.spaces[name]['symbol']: name
                        for name in self.spaces}
        # Construct a list of characters that the diecase contains
        spaces = [space for space in spaces_names]
        available_characters = COMMANDS + matrices + spaces
        # Characters which will be skipped
        ignored = ('\n',)
        # What if char in text not present in diecase? Hmmm...
        for index, _ in enumerate(input_text):
            if skip_steps:
                # Skip it, decrease counter, yield nothing
                skip_steps -= 1
                continue
            for i in range(self.ligatures, 0, -1):
                # Start from longest, end with shortest
                try:
                    char = input_text[index:index+i]
                    skip_steps = i - 1
                    if char not in ignored and char in available_characters:
                        # Try to look it up in spaces
                        yield spaces_names.get(char, char)
                        # End on first (largest) combination found
                        break
                except IndexError:
                    # Cannot generate a ligature (no more characters in input)
                    # Iterate further
                    pass
            # Should add a custom character definition here...


class OutputData(object):
    """Responsible for justifying, adding comments and outputting the
    ribbon"""
    def __init__(self):
        self.comments = False
        self.buffer = []
        self.ribbon = typesetting_data.Ribbon()

    def _justify_line(self, mode=1):
        """justify_line(mode=1)

        Justify the row; applies to all alignment routines.
        This function supports various modes:
        0: justification only by variable spaces
        1: filling the line with one block of fixed spaces, then dividing
           the remaining units among variable spaces
        2: as above but with two blocks of fixed spaces
        3, 4... - as above but with 3, 4... blocks of fixed spaces
        Add fixed spaces only if mode is greater than 0
        """
        # Add as many fixed spaces as we can
        # Don't exceed the line length (in units) specified in setup!
        # Predict if the increment will exceed it or not
        UI.debug_info('Justifying line...')
        fill_spaces_number = 0
        space_units = self.spaces['fixed']['units']
        # Determine if we have to add any spaces (otherwise - skip the loop)
        result_length = self.current_units + mode * space_units
        # Start with no spaces
        fill_spaces = []
        while mode and result_length < self.unit_line_length:
            # Add units
            self.current_units += mode * self.spaces['fixed']['units']
            # Add a mode-dictated number of fill spaces
            fill_spaces_number += mode
            # Determine and add the space code to the line
            space = list(self.spaces['fixed']['code'])
            space.append('Fixed space %i units wide'
                         % self.spaces['fixed']['units'])
            space = tuple(space)
            # Add as many spaces as needed
            fill_spaces = [space for i in range(fill_spaces_number)]
            # Update resulting length
            result_length = self.current_units + mode * space_units
        # The remaining units must be divided among the justifying spaces
        # Determine the unit width
        remaining_units = self.unit_line_length - self.current_units
        var_space_units = remaining_units / self.current_line_var_spaces
        # Return the space chunk (that will be appended at the beginning
        # or the end of the line, or both) and unit width
        return (fill_spaces, var_space_units)

    def _start_new_line(self, var_space_units):
        """Starts a new line during typesetting"""
        # Pass the unit width to the justify method later on
        self.line_buffer.append(('newline', var_space_units, 'New line'))
        self.buffer.extend(self.line_buffer)
        self.line_buffer = []
        self.current_units = 0
        self.current_line_var_spaces = 0

    def _align_left(self):
        """Aligns the previous chunk to the left."""
        UI.debug_info('Aligning line to the left...')
        (spaces, var_space_units) = self._justify_line(mode=1)
        self.line_buffer.extend(spaces)
        self._start_new_line(var_space_units)

    def _align_right(self):
        """Aligns the previous chunk to the right."""
        UI.debug_info('Aligning line to the right...')
        (spaces, var_space_units) = self._justify_line(mode=1)
        self.line_buffer = spaces + self.line_buffer
        self._start_new_line(var_space_units)

    def _align_center(self):
        """Aligns the previous chunk to the center."""
        UI.debug_info('Aligning line to the center...')
        (spaces, var_space_units) = self._justify_line(mode=2)
        # Prepare the line: (content, wedge_positions)
        self.line_buffer = spaces + self.line_buffer + spaces
        self._start_new_line(var_space_units)

    def _align_both(self):
        """Aligns the previous chunk to both edges and ends the line."""
        UI.debug_info('Aligning line to both edges...')
        (_, var_space_units) = self._justify_line(mode=0)
        self._start_new_line(var_space_units)


class Translator(object):
    """Typesetting class"""
    def __init__(self):
        self.settings = {'unit_line_length': 0, 'current_units': 0,
                         'ligatures': 3, 'current_line_var_spaces': 0,
                         'custom_characters': [], 'current_unit_correction': 0,
                         'main_style': 'roman', 'current_style': 'roman',
                         'current_alignment': 'align_left'}
        # Set up the spaces
        self.spaces = {'var': {'high': False, 'units': 4, 'symbol': ' ',
                               'code': 'GS2'},
                       'fixed': {'high': False, 'units': 9, 'code': 'G5',
                                 'symbol': '_'},
                       'nb': {'high': False, 'units': 9, 'code': 'G5',
                              'symbol': '~'},
                       'quad': {'high': False, 'units': 18, 'code': 'O15',
                                'symbol': '\t'}}
        self.buffers = {'line': [], 'work': []}
        self.compose = self.auto_compose
        self.current_alignment = self._align_left
        # By default use automatic typesetting - less user control
        self.manual_control = False

    def _choose_alignment(self):
        """Lets the user choose the text alignment in line or column."""
        options = {'L': 'align_left',
                   'C': 'align_center',
                   'R': 'align_right',
                   'B': 'align_both'}
        message = ('Default alignment: [L]eft, [C]enter, [R]ight, [B]oth? ')
        self.settings['current_alignment'] = UI.simple_menu(message, options)

    def translate(self, char):
        """translate:

        Translates the character to a combination of Monotype signals,
        applying single or double justification whenever necessary.
        Returns an unit value of the character, so that it can be added
        to current line's unit length.
        """
        try:
            # Is that a command?
            self.typesetting_commands[char]()
            # Return 0 unit increment
            return 0
        except KeyError:
            # If not, then continue
            pass
        # Detect any spaces and quads
        for name in self.spaces:
            if char == self.spaces[name]['symbol']:
                units = self.spaces[name]['units']
                self.line_buffer.append((name, units,
                                         self.spaces[name]['desc']))
                if name == 'var':
                    self.current_line_var_spaces += 1
                return units
        # Space not recognized - so this is a character.
        # Get the matrix data: [char, style, column, row, units]
        # First try custom-defined characters (overrides)
        # If empty - try diecase layout
        matches = ([mat for mat in self.custom_characters
                    if mat[0] == char and self.current_style in mat[1]] or
                   [mat for mat in self.diecase_layout if mat[0] == char and
                    self.current_style in mat[1]])
        while not matches:
            # Nothing found in the diecase for this character?
            # Define it then yourself!
            matrix = []
            UI.display('Enter the position for character %s, style: %s'
                       % (char, self.current_style))
            row = UI.enter_data('Column? ').upper
            column = UI.enter_data('Row? ', int)
            matrix = [mat for mat in self.diecase_layout
                      if mat[2] == column and mat[3] == row]
        if len(matches) == 1:
            matrix = matches[0]
        elif len(matches) > 1:
            options = dict(enumerate(matches, start=1))
            matrix = UI.simple_menu('Choose a matrix for the character %s, '
                                    'style: %s' % (char, self.current_style),
                                    options)
            self.custom_characters.append(matrix)
        # If char units is the same as the row units, no correction is needed
        # Get coordinates
        column = matrix[2]
        row = matrix[3]
        # Combination to be cast
        combination = column + str(row)
        normal_unit_width = matrix[4]
        # Add or subtract current unit correction
        char_units = normal_unit_width + self.unit_correction
        # Finally add the translated character to output buffer
        self.line_buffer.append((combination, char_units, char))
        # Return the character's unit width
        return char_units

    def _get_space_code(self, desired_unit_width, high_space=False):
        """get_space_code:

        Gets coordinates for the space of a desired unit width.
        First, looks for spaces of matching width. If it fails,
        uses unit-shift (if available) for a broader choice of spaces,
        and if this fails too, applies justification with the S-needle.

        Returns (space_code, wedge_positions).
        Adds 'S' to the space code if justification is applied. In this case,
        wedge_positions are (pos0075, pos0005) - min. (1, 1), max (15, 15).
        Wedge positions are (None, None) if justification is not applied.
        """
        # Space style: " " for low space and "_" for high space
        space_symbols = {'□': False, ' ': False, '_': True, '▣': True}
        available_spaces = [sp for sp in self.diecase_layout if
                            sp[0] == space_symbols[high_space]]
        spaces = []
        # Gather non-shifted spaces
        for space in available_spaces:
            # Get the unit value for this space
            column = space[2]
            row = space[3]
            if row < 16:
                space_unit_width = self.units[row]
                space_code = column + str(row)
                spaces.append((space_unit_width, space_code))
        # Try matching a first available space
        for (space_unit_width, space_code) in spaces:
            unit_difference = desired_unit_width - space_unit_width
            if 0 <= unit_difference < 10:
                # Unit correction: add max 10
                return space_code
        else:
            # Return the space that is narrower and units can be added
            return False

    def _configure_spaces(self):
        """Chooses the spaces that will be used in typesetting."""
        # To make lines shorter
        enter = UI.enter_data_or_blank
        enter_type = UI.enter_data_or_blank
        get_space_code = self._get_space_code
        y_n = UI.confirm
        # List available spaces
        # Matrix in layout is defined as follows:
        # (character, (style1, style2...)) : (column, row, unit_width)
        var_units_prompt = 'How many units min. (default: 4)? '
        var_symbol_prompt = ('Variable space symbol in text file? '
                             '(default: " ") ? ')
        fixed_units_prompt = 'How many units (default: 9)? '
        fixed_symbol_prompt = ('Fixed space symbol in text file? '
                               '(default: "_") ? ')
        nbsp_units_prompt = 'How many units (default: 9)? '
        nbsp_symbol_prompt = ('Non-breaking space symbol in text file? '
                              '(default: "~") ? ')
        quad_symbol_prompt = ('Em-quad symbol in text file? '
                              '(default: "\\t") ? ')
        # Choose spaces
        spaces = {}
        # Variable space (justified) - minimum units
        # Ask if low or high and save the choice
        variable_space = y_n('Variable: is a high space? ')
        spaces['var']['high'] = variable_space
        # Ask for minimum number of units of given set for the variable space
        spaces['var']['units'] = enter_type(var_units_prompt, int) or 4
        # Variable space code will be determined during justification
        # Don't do it now yet - we don't know the wedge positions
        spaces['var']['code'] = 'GS2'
        # Ask for the symbol representing the space in text
        spaces['var']['symbol'] = enter(var_symbol_prompt) or ' '
        # Fixed space (allows line-breaking)
        # Ask if low or high and save the choice
        spaces['fixed']['high'] = y_n('Fixed: is a high space? ')
        # Ask for unit-width of this space
        spaces['fixed']['units'] = enter_type(fixed_units_prompt, int) or 9
        # Determine fixed space code
        spaces['fixed']['code'] = get_space_code(spaces['fixed']['high'],
                                                 spaces['fixed']['units'])
        # Ask for the symbol representing the space in text
        spaces['fixed']['symbol'] = enter(fixed_symbol_prompt) or '_'
        # Non-breaking space
        # Ask if low or high and save the choice
        spaces['nb']['high'] = y_n('Non-breaking: is a high space? ')
        # Ask for unit-width of this space
        spaces['nb']['units'] = enter_type(nbsp_units_prompt, int) or 9
        # Determine non-breaking space code
        spaces['nb']['code'] = get_space_code(spaces['nb']['high'],
                                              spaces['nb']['units'])
        # Ask for the symbol representing the space in text
        spaces['nb']['symbol'] = enter(nbsp_symbol_prompt) or '~'
        # Set up the quad code arbitrarily
        spaces['quad']['code'] = 'O15'
        spaces['quad']['high'] = False
        spaces['quad']['units'] = 18
        # Em quad symbol choice
        spaces['quad']['symbol'] = enter(quad_symbol_prompt) or '\t'
        # Finalize setup
        self.spaces = spaces

    def manual_compose(self):
        """Reads text fragments from input, then composes them, and justifies
        to a specified line length. Adds codes to the buffer.
        Text fragments is a list of tuples: ((text1, style1), ...)
        """

    def auto_compose(self):
        """Composes text automatically, deciding when to end the lines."""
        # Start with the empty work buffer
        self.buffer = []
        self.line_buffer = []
        self.current_units = 0
        finished = False
        # Entering a line
        while not finished:
            # Keep looping over all characters and lines
            # Try to fill the line and not hyphenate
            while self.current_units < self.unit_line_length - 50:
                # Get the character from input
                try:
                    character = next(self.text_source)
                # Translate the character (add it to buffer),
                # get unit width for the character from function's retval
                    self.current_units += self.translate(character)
                except StopIteration:
                    finished = True
                    break
            # Line composed now, align and justify it
            UI.debug_info('Line finished. Now aligning...')
            self.current_alignment()
        # Now we're done typesetting
        UI.pause('Typesetting finished!')
        return True


    def justify(self):
        """justify:

        Reads the codes buffer backwards and checks if character needs
        unit width correction (unit_difference != 0) - if so, calls
        calculate_wedges to get the 0075 and 0005 wedge positions for the char.
        If corrections are needed, applies single justification.
        If "newline" combination is found, sets the wedges for a line.
        Outputs the ready sequence.
        """
        line_wedge_positions = (3, 8)
        current_wedge_positions = (3, 8)
        var_space_high = self.spaces['var']['high']
        while self.buffer:
            # Take the last combination off
            (combination, unit_difference, comment) = self.buffer.pop()
            # If comments are enabled, and there is one, append it
            # Else, append an empty string
            comment = self.comments and comment and ' // ' + comment or ''
            # New line - use double justification;
            # instead of unit difference between row and mat, we have
            # unit width of a variable space
            if combination == 'newline':
                # Variable space parameters: (var_space_code, wedge_positions)
                var_space_code = self._get_space_code(unit_difference,
                                                      var_space_high) + 'S'
                current_wedge_positions = line_wedge_positions
                self.double_justification(line_wedge_positions)
            # Justifying space - if wedges were set to different positions,
            # reset them to line justification positions
            elif combination == 'var_space':
                if current_wedge_positions != line_wedge_positions:
                    # Set the line justification
                    current_wedge_positions = line_wedge_positions
                    self.single_justification(current_wedge_positions)
                self.output_buffer.append(self.spaces['var']['code'] + comment)
            elif combination == 'fixed_space':
                pass
            elif not unit_difference:
                # No corrections needed
                self.output_buffer.append(combination + comment)
            if wedge_positions != current_wedge_positions:
                # Correction needed - determine if wedges are already set
                self.single_justification(wedge_positions)
                current_wedge_positions = wedge_positions
                self.output_buffer.append(combination + comment)
        return self.output_buffer

    def single_justification(self, wedge_positions):
        """single_justification:

        Single justification: the caster sets 0005, then 0075 + pump resumes.
        Function adds a 0075-b, then 0005-a combination to buffer.
        Supports both normal (0075, 0005, 0005+0075) and alternate
        (NK, NJ, NKJ) justification modes. Adds an extra "S" signal
        to assist setting the 0005 and 0075 levers in place (some machines
        have a problem with that).
        """
        # This function is used during backwards-parsing and converting
        # the combinations sequence after typesetting
        # Codes are placed in the sequence that will be read by the caster
        # So, we get 0005 first, then 0075
        (pos0075, pos0005) = wedge_positions
        # Inform the user about wedge positions
        comment1 = ' // Pump off; setting 0005 to ' + str(pos0005)
        comment2 = ' // Setting 0075 wedge to ' + str(pos0075) + ', pump on...'
        # Add 0005-N-J-S-pos0005 first:
        self.output_buffer.append(str(pos0005) + 'NJS 0005' + comment1)
        # Add 0075-N-K-S-pos0075 next:
        self.output_buffer.append(str(pos0075) + 'NKS 0075' + comment2)
        return True

    def double_justification(self, wedge_positions):
        """double_justification:

        Double justification: the caster sets 0005+0075 and puts the line
        to the galley, then sets 0075 and pump resumes.
        Function adds a 0075-b, then 0005-a combination to buffer.
        Supports both normal (0075, 0005, 0005+0075)
        and alternate (NK, NJ, NKJ) justification modes.
        Adds an "S" signal also - to help delivering the suitable force
        to the 0005 and 0075 levers (some machines have a problem with that).
        """
        (pos0075, pos0005) = wedge_positions
        # Inform the user about the wedge positions
        comment1 = ' // Line to the galley, setting 0005 to ' + str(pos0005)
        comment2 = ' // Setting 0075 to ' + str(pos0075) + ', pump on...'
        # Add 0005-N-J-S-pos0005 first:
        self.output_buffer.append(str(pos0005) + 'NKJS 0005 0075' + comment1)
        # Add 0075-N-K-S-pos0075 next:
        self.output_buffer.append(str(pos0075) + 'NKS 0075' + comment2)
        return True


def parse_combination(combination):
    """Simple function that gets the row and column from combination string."""
    combination = str(combination).upper()
    rows = []
    # Look for column numbers from earliest to latest
    column_numbers = ['NI', 'NL'] + list('ABCDEFGHIJKLMNO')
    for num in column_numbers:
        if num in combination:
            column = num
            break
    else:
        # No signal for column means we go all the way to "O" pin
        column = 'O'
    # Now look for a row number
    # We can't look for them in ascending order though
    for num in range(16, 0, -1):
        if str(num) in combination:
            combination = combination.replace(str(num), '')
            rows.append(num)
    # Get the minimum number
    print(rows)
    try:
        row = min(rows)
    except ValueError:
        # This happens if no rows found - default to 15
        row = 15
    return (column, row)



# -*- coding: utf-8 -*-
"""Global settings for rpi2caster"""

from . import text_ui

# User interface we use (text, web etc.)
UI = text_ui
# Conffile path
CONFIG_PATH = '/etc/rpi2caster.conf'
# SQLite3 database path
DATABASE_PATH = '/var/local/rpi2caster/rpi2caster.db'
# Comment symbols for parsing
COMMENT_SYMBOLS = ['**', '*', '//', '##', '#']
# GPIO settings
# Inputs for the casting interface
EMERGENCY_STOP_GPIO = 22
SENSOR_GPIO = 17
# Default hardware interface output data
MCP0 = 0x20
MCP1 = 0x21
PIN_BASE = 65



# -*- coding: utf-8 -*-
"""Wedge data model for rpi2caster"""
from itertools import zip_longest
# Default user interface
from .global_settings import UI
# Constants for known normal wedge unit values
from . import wedge_unit_values as wu


class Wedge(object):
    """Default S5-12E wedge"""
    def __init__(self):
        self.series = '5'
        self.set_width = 12
        self.is_brit_pica = True
        self.units = wu.S5

    def __repr__(self):
        return self.name

    @property
    def pica(self):
        """Get the pica value for the wedge. Can be .1667" (old British)
        or .166" (new British; American). The .1667" was commonly used with
        wedges made for European markets (wedge designation series-setE).
        Curiously, the old British pica is the same as modern DTP pica."""
        return self.is_brit_pica and 0.1667 or 0.166

    @property
    def em_width(self):
        """Get the wedge em width (i.e. 18 units of wedge's set) in inches"""
        return self.pica * self.set_width / 12

    @property
    def name(self):
        """Gets a wedge name - for example S5-12.25E.
        S (for stopbar) is prepended by convention.
        E is appended whenever the wedge is based on pica = .1667".
        """
        # Avoid displaying ".0" in non-fractional set widths
        if self.set_width % 1 == 0:
            set_width = int(self.set_width)
        else:
            set_width = self.set_width
        name = 'S%s-%s%s' % (self.series, set_width, self.is_brit_pica * 'E')
        return name

    @name.setter
    def name(self, wedge_name=None):
        """Defines a wedge based on designation.
        Recognizes many new-style (numeric) names, like S5-12 or 5-12,
        and some old-style (alphabetic) names, like AK, BO etc.
        For unknown wedges, the user has to enter the values manually."""
        # Ask for wedge name and set width as it is written on the wedge
        prompt = ('\nSome old-style wedge designations:\n\n' +
                  '\n'.join(['\t'.join(group)
                             for group in grouper(wu.ALIASES, 3, '')]) +
                  '\n\nIf you have one of those, enter number (like S5-xx.yE).'
                  '\n\nWedge designation? (blank to choose default S5-12): ')
        wedge_name = wedge_name or UI.enter_data_or_blank(prompt) or 'S5-12'
        # For countries that use comma as decimal delimiter, convert to point:
        wedge_name = wedge_name.replace(',', '.').upper().strip()
        # Check if this is an European wedge
        # (these were based on pica = .1667" )
        is_brit_pica = wedge_name.endswith('E')
        # Away with the initial S, final E and any spaces before and after
        # Make it work with space or dash as delimiter
        wedge = wedge_name.strip('SE ').replace('-', ' ').split(' ')
        try:
            (series, set_width) = wedge
        except ValueError:
            series = wedge[0]
            set_width = None
        # Now get the set width - ensure that it is float divisible by 0.25
        while True:
            try:
                set_width = float(set_width)
                if not set_width % 0.25:
                    # Correct value, finish here
                    break
                else:
                    # Not divisible by 0.25 => wrong value, re-enter
                    raise ValueError
            except (TypeError, ValueError):
                prompt = ('Enter the set width as a decimal fraction '
                          'divisible by 0.25 - e.g. 12.25: ')
                set_width = UI.enter_data(prompt, float)
        # We have the wedge name, so we can look the wedge up in known wedges
        # (no need to enter the unit values manually)
        units = wu.UNITS.get(series, None)
        while not units:
            # Enter manually if not found:
            prompt = ('Enter the wedge unit values for rows 1...15 '
                      'or 1...16, separated by commas.\n')
            # Now we need to be sure that all whitespace is stripped,
            # and the value written to database is a list of integers
            try:
                units = [int(i.strip())
                         for i in UI.enter_data(prompt).split(',')]
                # Display warning if the number of steps is not
                # 15 or 16 (15 is most common, 16 was used for HMN and KMN).
                if not 15 <= len(units) <= 16:
                    UI.display('Wedge must have 15 or 16 steps - enter again.')
                    units = None
            except ValueError:
                UI.display('Incorrect value - enter the values again.')
                units = None
        # Finally set the collected data
        self.series = series
        self.set_width = set_width
        self.is_brit_pica = is_brit_pica
        self.units = units

    @property
    def parameters(self):
        """Gets a list of parameters"""
        return [(self, 'Wedge designation'),
                ('%.4f"' % self.pica, 'Pica base is'),
                (' '.join([str(x) for x in self.units if x]),
                 'Row unit values for this wedge')]

    @property
    def units(self):
        """Gets the unit values for the wedge's rows"""
        units = self.__dict__.get('_units', None) or wu.S5
        # Add 0 at the beginning so that the list can be indexed with
        # row numbers - i.e. units[1] for row 1
        if units[0] is not 0:
            units = [0] + units
        # Ensure that wedge has 0 + 16 steps - i.e. length is 17
        # Keep adding the last value until we have enough
        while len(units) < 17:
            units.append(units[-1])
        return units

    @units.setter
    def units(self, units):
        """Sets the wedge unit values"""
        self.__dict__['_units'] = units


class SelectWedge(Wedge):
    """Selects a wedge automatically or manually"""
    def __init__(self, wedge_name=None, wedge_series=None, set_width=None):
        super().__init__()
        self.name = wedge_name or (wedge_series and set_width and
                                   '%s-%s' % (wedge_series, set_width))


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)



# -*- coding: utf-8 -*-
"""Contains custom exceptions for the rpi2caster package. """


class ExceptionRoot(Exception):
    """Superclass for all custom exceptions defined for rpi2caster"""


class DatabaseOops(ExceptionRoot):
    """Superclass for database exceptions"""


class ConfigOops(ExceptionRoot):
    """Config-related exceptions"""


class CastingOops(ExceptionRoot):
    """Superclass for errors in the casting process"""


class MissingDependency(ExceptionRoot):
    """Raised when dependencies are not satisfied."""
    pass


class UIOops(ExceptionRoot):
    """Exceptions raised and caught by the UI"""
    pass


class TypesettingOops(ExceptionRoot):
    """Exceptions raised in the typpesetting process"""


class CastingAborted(CastingOops):
    """Raised when casting is aborted due to machine stop."""
    pass

class CastingFinished(CastingOops):
    """Raised when casting is finished normally because ribbon runs out."""
    pass

class MachineStopped(CastingOops):
    """Raised by the caster control routines, when the machine stops."""
    pass


class EmergencyStop(CastingOops):
    """Raised when user pushes the emergency stop button to abort."""
    pass


class ReturnToMenu(UIOops):
    """Raised when a method wants to exit to menu."""
    pass


class MenuLevelUp(UIOops):
    """Raised in submenus, to go up to a higher level"""
    pass


class ExitProgram(UIOops):
    """Raised when user decides to exit program."""
    pass


class ChangeParameters(UIOops):
    """Raised when user decides to change parameters in a current routine. """
    pass


class Row16(ChangeParameters):
    """Raised when row 16 found in parsing - turn on HMN/KMN/unit-shift"""
    pass


class ConfigFileUnavailable(ConfigOops):
    """Raised when a config file does not exist or cannot be read."""
    pass


class WrongConfiguration(ConfigOops):
    """Raised when a parameter is not configured correctly."""
    pass


class NotConfigured(ConfigOops):
    """Raised when a section or parameter is missing."""


class NoMatchingData(DatabaseOops):
    """Raised when a database query yields no matches."""
    pass


class DatabaseQueryError(DatabaseOops):
    """Raised by the low-level database module as a general exception
    if a SQL query fails to execute.
    """
    pass


class MatrixNotFound(TypesettingOops):
    """Raised when the typesetting program could not find a desired
    character in the diecase layout."""
    pass


class DuplicateError(TypesettingOops):
    """DuplicateError

    Raised when trying to add a duplicate item:
    matrix (with identical coordinates), diecase (with identical ID).
    """
    pass

# Little helper functions that raise the exceptions


def return_to_menu():
    """Wrapper for raising ReturnToMenu where a function is needed"""
    raise ReturnToMenu


def abort_casting():
    """Wrapper for raising CastingAborted where a function is needed"""
    raise CastingAborted


def menu_level_up():
    """Wrapper for raising MenuLevelUp where a function is needed"""
    raise MenuLevelUp


def exit_program():
    """Wrapper for raising ReturnToMenu where a function is needed"""
    raise ExitProgram


def change_parameters():
    """Wrapper for raising ChangeParameters where a function is needed"""
    raise ChangeParameters



# -*- coding: utf-8 -*-
"""
Kernel SysFS-based input drivers for photocell and machine cycle sensor.
"""
import io
# Essential for polling the sensor for state change:
import select
# Debounce timers need this
from time import time
# Constants shared between modules
from .global_settings import UI, SENSOR_GPIO
# Custom exceptions
from .exceptions import MachineStopped
# Caster prototype
from .monotype import SimulationSensor


class SysfsSensor(SimulationSensor):
    """Optical cycle sensor using kernel sysfs interface"""
    def __init__(self, gpio=SENSOR_GPIO):
        super().__init__()
        self.gpio = gpio
        self.signals = None
        self.last_state = False
        self.value_file_obj = None
        self.name = 'Kernel SysFS interface for photocell sensor GPIO'
        try:
            self.value_file = configure_sysfs_interface(gpio)
        except TypeError:
            self.value_file = '/dev/null'
            self.edge_file = '/dev/null'

    @property
    def parameters(self):
        """Gets a list of parameters"""
        return [(self.name, 'Sensor driver'),
                (self.gpio, 'GPIO number'),
                (self.value_file, 'Value file path'),
                (self.edge_file, 'Edge file path')]

    def wait_for(self, new_state, timeout=5, *_):
        """
        Waits until the sensor is in the desired state.
        new_state = True or False.
        timeout means that if no signals in given time, raise MachineStopped.
        force_cycle means that if last_state == new_state, a full cycle must
        pass before exit.
        Uses software debouncing set at 50ms
        """
        def get_state():
            """Reads current input state"""
            gpiostate.seek(0)
            # File can contain "1\n" or "0\n"; convert it to boolean
            return bool(int(gpiostate.read().strip()))

        # Set debounce time to now
        debounce = time()
        # Prevent sudden exit in the midst of machine cycle
        with io.open(self.value_file, 'r') as gpiostate:
            if get_state() == new_state:
                self.last_state = not new_state
        with io.open(self.value_file, 'r') as gpiostate:
            signals = select.epoll()
            signals.register(gpiostate, select.POLLPRI)
            while self.last_state != new_state:
                if signals.poll(timeout):
                    state = get_state()
                    if time() - debounce > 0.05:
                        self.last_state = state
                    debounce = time()
                else:
                    raise MachineStopped


def configure_sysfs_interface(gpio):
    """configure_sysfs_interface(gpio):

    Sets up the sysfs interface for reading events from GPIO
    (general purpose input/output). Checks if path/file is readable.
    Returns the value and edge filenames for this GPIO.
    """
    # Set up an input polling file for machine cycle sensor:
    gpio_sysfs_path = '/sys/class/gpio/gpio%s/' % gpio
    gpio_value_file = gpio_sysfs_path + 'value'
    gpio_edge_file = gpio_sysfs_path + 'edge'
    # Check if the GPIO has been configured - file is readable:
    try:
        with io.open(gpio_value_file, 'r'):
            pass
        # Ensure that the interrupts are generated for sensor GPIO
        # for both rising and falling edge:
        with io.open(gpio_edge_file, 'r') as edge_file:
            if 'both' not in edge_file.read():
                UI.display('%s: file does not exist, cannot be read, '
                           'or the interrupt on GPIO %i is not set '
                           'to "both". Check the system configuration.'
                           % (gpio_edge_file, gpio))
    except (IOError, FileNotFoundError):
        UI.display('%s : file does not exist or cannot be read. '
                   'You must export the GPIO no %s as input first!'
                   % (gpio_value_file, gpio))
    else:
        return gpio_value_file



# -*- coding: utf-8 -*-
"""Module containing functions for diecase retrieval and parsing,
diecase storing, parameter searches.
"""
# File operations
import io
import os
import csv

# Some functions raise custom exceptions
from . import exceptions as e
# Wedge operations for several matrix-case management functions
from . import wedge_data
# Constants module
from . import constants as c
# Parsing module
from . import parsing as p
# Database backend
from . import database
# User interface
from .global_settings import UI
DB = database.Database()


class Diecase(object):
    """Diecase: matrix case attributes and operations"""
    def __init__(self):
        self.diecase_id = ''
        self.typeface = ''
        self.wedge = wedge_data.Wedge()
        self.matrices = generate_empty_layout(15, 17)

    def __iter__(self):
        return iter(self.matrices)

    def __next__(self):
        yield from self.matrices

    def __repr__(self):
        return self.diecase_id

    def __bool__(self):
        return bool(self.diecase_id)

    def __getitem__(self, char_and_style):
        try:
            char, style = char_and_style
        except ValueError:
            # Must work if we give it a character
            char = char_and_style
            if len(self.styles) == 1:
                style = self.styles[0]
            else:
                style = 'roman'
        return self.get_matrix(char, style)

    def show_layout(self):
        """Shows the diecase layout"""
        UI.display_diecase_layout(self)
        UI.pause()

    def edit_layout(self):
        """Edits a layout and asks if user wants to save changes"""
        # Edit the layout
        UI.edit_diecase_layout(self)

    def get_matrix(self, char=None, style='roman'):
        """Chooses a matrix automatically or manually (if multiple matches),
        allows to specify matrix data manually if no matches found"""
        if char in ' _':
            style = ''
        candidates = [mat for mat in self.matrices
                      if mat.char == char and
                      (not any(mat.styles) or style in mat.styles)]
        if not candidates:
            matrix = Matrix(char or '', [style])
            matrix.diecase = self
            if char:
                UI.display('Enter matrix data for character: %s' % char)
            prompt = 'Combination? (default: G5): '
            code_string = UI.enter_data_or_blank(prompt).upper() or 'G5'
            combination = p.parse_signals(code_string)
            # Got a list of signals
            matrix.row = p.get_row(combination)
            matrix.column = p.get_column(combination)
            # Default unit width value = determine based on wedge
            units = self.wedge.units[matrix.row]
            prompt = 'Unit width? (default: %s): ' % units
            matrix.units = abs(UI.enter_data_or_blank(prompt, int) or units)
            return matrix
        elif len(candidates) == 1:
            return candidates[0]
        else:
            pr_char = {' ': 'low space', '_': 'high space'}.get(char, char)
            UI.display_header('Multiple matrices for %s %s' % (style, pr_char))
            # Show a menu with multiple candidates
            hdr = (['Index'.ljust(10), 'Char'.ljust(10), 'Styles'.ljust(30),
                    'Column'.ljust(10), 'Row'.ljust(10), 'Units'.ljust(10)])
            mats = {i: mat for i, mat in enumerate(candidates, start=1)}
            prompt = 'Choose matrix (leave blank to enter manually): '
            choice = 0
            UI.display(''.join(hdr))
            for i, mat in mats.items():
                record = [str(i).ljust(10), mat.char.ljust(10),
                          ', '.join(mat.styles).ljust(30),
                          mat.column.ljust(10), str(mat.row).ljust(10),
                          str(mat.units).ljust(10)]
                UI.display(''.join(record))
            choice = UI.enter_data_or_blank(prompt, int)
            matrix = mats.get(choice, Matrix(char, [style]))
            matrix.diecase = self
            return matrix

    def decode_matrix(self, code):
        """Finds the matrix based on the column and row in layout"""
        return [mat for mat in self.matrices if mat == code.upper()][0]

    def import_layout(self):
        """Imports a layout from file"""
        # Load the layout from file
        submitted_layout = import_layout_file()
        if not submitted_layout:
            UI.pause('File does not contain a proper layout!')
            return False
        # Update the empty layout with characters read from file
        # record = (char, styles, column, row, units)
        try:
            self.matrices = generate_empty_layout()
            for matrix in self.matrices:
                for (char, styles, column, row, units) in submitted_layout:
                    if matrix.column == column and matrix.row == row:
                        matrix.char = char
                        matrix.styles = styles
                        matrix.units = units
        except (TypeError, ValueError):
            UI.pause('Cannot process the uploaded layout!')
            return False

    def export_layout(self):
        """Exports the matrix case layout to file."""
        filename = os.path.expanduser('~') + '/%s.csv' % self
        with io.open(filename, 'a') as output_file:
            csv_writer = csv.writer(output_file, delimiter=';', quotechar='"',
                                    quoting=csv.QUOTE_ALL)
            csv_writer.writerow(['Character', 'Style(s)', 'Column', 'Row',
                                 'Unit width'])
            for matrix in self.matrices:
                (char, styles, column, row, units) = matrix.record
                csv_writer.writerow([char, ', '.join(list(styles)),
                                     column, row, units])
        UI.pause('File %s successfully saved.' % filename)
        return True

    def check_missing_characters(self, input_string='', style='roman'):
        """Enter the string and parse the diecase to see if any of the
        specified characters are missing."""
        input_string = input_string or UI.enter_data('Text to check: ')
        charset = {char for char in input_string}
        chars_found = {mat.char for mat in self for char in charset
                       if char == mat.char and
                       (char in ' _' or style in mat.styles)}
        missing = sorted(charset.difference(chars_found))
        if missing:
            UI.display('Missing mats for %s: %s' % (style, ', '.join(missing)))
        else:
            UI.display('The diecase has all characters we need.')

    def set_diecase_id(self, diecase_id=None):
        """Sets a diecase ID"""
        prompt = 'Diecase ID? (leave blank to exit) : '
        diecase_id = (diecase_id or UI.enter_data_or_blank(prompt) or self)
        # Ask if we are sure we want to update this
        # if self.diecase_id was set earlier
        if not self.diecase_id or UI.confirm('Apply changes?', default=False):
            self.diecase_id = diecase_id
            return True

    def set_typeface(self, typeface=None):
        """Sets the type series, size and typeface name"""
        prompt = 'Typeface (series, size, name): '
        typeface = typeface or UI.enter_data_or_blank(prompt) or self.typeface
        if not self.typeface or UI.confirm('Apply changes?', default=False):
            self.typeface = typeface
            return True

    def assign_wedge(self, wedge_name=None):
        """Assigns a wedge (from database or newly-defined) to the diecase"""
        self.wedge = wedge_data.SelectWedge(wedge_name)

    def save_to_db(self):
        """Stores the matrix case definition/layout in database"""
        try:
            DB.add_diecase(self)
            UI.pause('Data saved successfully.')
            return True
        except e.DatabaseQueryError:
            UI.pause('Cannot save the diecase!')

    def delete_from_db(self):
        """Deletes a diecase from database"""
        ans = UI.confirm('Are you sure?', default=False)
        if ans and DB.delete_diecase(self):
            UI.display('Matrix case deleted successfully.')
            return True

    def check_db(self):
        """Checks if the diecase is registered in database"""
        try:
            DB.get_diecase(self.diecase_id)
            return True
        except (e.DatabaseQueryError, e.NoMatchingData):
            return False

    @property
    def layout(self):
        """Gets a diecase layout as a list of lists"""
        return [matrix.record for matrix in self.matrices]

    @layout.setter
    def layout(self, layout):
        """Translates the layout to a list of matrix objects"""
        self.matrices = [Matrix(char, styles, (column, row), units)
                         for (char, styles, column, row, units) in layout]

    @property
    def matrices(self):
        """Gets a list of Matrix objects"""
        return self.__dict__.get('_matrices')

    @matrices.setter
    def matrices(self, matrices):
        """Sets a list of Matrix objects, used for storing the layout"""
        if matrices:
            for matrix in matrices:
                matrix.diecase = self
            self.__dict__['_matrices'] = matrices

    @property
    def parameters(self):
        """Gets a list of parameters"""
        return [(self.diecase_id, 'Diecase ID'),
                (self.typeface, 'Typeface'),
                (self.wedge, 'Assigned wedge')]

    @property
    def styles(self):
        """Parses the diecase layout and gets available typeface styles.
        Returns a list of them."""
        return list({style for mat in self for style in mat.styles if style})

    def manipulation_menu(self):
        """A menu with all operations on a diecase"""
        def clear_layout():
            """Generates a new layout for the diecase"""
            self.matrices = generate_empty_layout()
        try:
            while True:
                UI.clear()
                UI.display_parameters({'Diecase data': self.parameters,
                                       'Wedge data': self.wedge.parameters})
                options = {'M': e.return_to_menu,
                           'T': self.set_typeface,
                           'W': self.assign_wedge,
                           'ID': self.set_diecase_id,
                           'E': self.edit_layout,
                           'N': clear_layout,
                           'V': self.show_layout,
                           'I': self.import_layout,
                           'X': self.export_layout,
                           'C': e.menu_level_up}
                messages = ['\nMatrix case manipulation:\n\n'
                            '[V]iew, [N]ew, [E]dit, [I]mport '
                            'or e[X]port layout\nAssign [W]edge, '
                            'change [T]ypeface or diecase [ID]\n\n']
                # Save to database needs a complete set of metadata
                required = {'Typeface': self.typeface,
                            'Diecase ID': self.diecase_id}
                missing = [item for item in required if not required[item]]
                messages.extend([item + ' is not set\n' for item in missing])
                if not missing:
                    options['S'] = self.save_to_db
                    messages.append('[S]ave diecase to database')
                # Check if it's in the database
                if self.check_db():
                    options['D'] = self.delete_from_db
                    messages.append(', [D]elete diecase from database')
                # Options constructed
                messages.append('\n[C] to choose/create another diecase, '
                                '[M] to exit to menu.')
                messages.append('\nYour choice: ')
                message = ''.join(messages)
                UI.simple_menu(message, options)()
        except e.MenuLevelUp:
            # Exit matrix case manipulation menu
            return True


class SelectDiecase(Diecase):
    """Initialize a diecase object with or without given ID.
    If possible, choose automatically. If this fails, choose manually,
    and if user chooses no diecase, an empty one will be initialized."""
    def __init__(self, diecase_id=None):
        super().__init__()
        # Diecases created with diecase_id will be set up automatically
        try:
            mcd = diecase_id and DB.get_diecase(diecase_id) or choose_diecase()
            (self.diecase_id, self.typeface, wedge_name, self.layout) = mcd
            # Assign the correct wedge
            self.wedge = wedge_data.SelectWedge(wedge_name)
        except (KeyError, TypeError, e.NoMatchingData, e.DatabaseQueryError):
            UI.display('Diecase choice failed. Using empty one instead.')


class Matrix(object):
    """A class for single matrices - all matrix data"""
    def __init__(self, char='', styles=('roman',),
                 coordinates=('O', 15), units=None):
        self.char = char
        self.styles = styles
        # Spaces (low and high) have every style
        if char in ' _':
            styles = []
        (self.column, self.row) = coordinates
        self.diecase = None
        self.units = units

    def __repr__(self):
        return self.code

    def row_units(self, alt_wedge=None):
        """Gets the row's unit width value for the diecase's wedge"""
        wedge = alt_wedge or self.diecase.wedge
        return wedge.units[self.row]

    @property
    def row(self):
        """Gets the row number"""
        return self.__dict__.get('_row', 5)

    @row.setter
    def row(self, value):
        """Gets the row number"""
        value = int(value)
        value = max(1, value)
        value = min(value, 16)
        self.__dict__['_row'] = value

    @property
    def parameters(self):
        """Gets all parameters for the matrix"""
        return [(self.char, 'Character'),
                (', '.join(list(self.styles)), 'Styles'),
                (self, 'Coordinates'), (self.units, 'Unit width')]

    @property
    def code(self):
        """Gets the matrix code"""
        return self.column + str(self.row)

    @property
    def units(self):
        """Gets the specific or default number of units"""
        return self.__dict__.get('_units', self.row_units())

    @units.setter
    def units(self, units):
        """Sets the unit width value"""
        self.__dict__['_units'] = units

    @property
    def record(self):
        """Returns a record suitable for JSON-dumping and storing in DB"""
        return [self.char, self.styles, self.column, self.row, self.units]

    @property
    def diecase(self):
        """Get a diecase or an empty one"""
        return self.__dict__.get('_diecase', Diecase())

    @diecase.setter
    def diecase(self, diecase):
        """Set a diecase"""
        if diecase is not None:
            self.__dict__['_diecase'] = diecase

    def wedge_positions(self, unit_correction=0, alt_wedge=None):
        """Calculates the 0075 and 0005 wedge positions for this matrix
        based on the diecase's default wedge or specified one."""
        wedge = alt_wedge or self.diecase.wedge
        diff = self.units + unit_correction - self.row_units(wedge)
        # 53 = neutral position where no corrections applied, i.e. 3/8
        # diff in units of given set; wedge pica = 0.166 or 0.1667
        steps_0005 = int(diff / wedge.pica * wedge.set_width * 0.25617) + 53
        # Upper limit: 15/15 => 15*15=225 + 15 = 240
        # (unsafe for casting from mats!)
        steps_0005 = min(steps_0005, 240)
        # Lower limit: 1/1 wedge positions => 15 + 1 = 16:
        steps_0005 = max(16, steps_0005)
        steps_0075 = 0
        while steps_0005 > 15:
            steps_0005 -= 15
            steps_0075 += 1
        # Got the wedge positions, return them
        return {'0075': steps_0075, '0005': steps_0005}

    def edit(self):
        """Edits the matrix data"""
        UI.edit_matrix(self)


def diecase_operations():
    """Matrix case operations menu for inventory management"""
    try:
        UI.display_header('Matrix case manipulation')
        while True:
            # Choose a diecase or initialize a new one
            SelectDiecase().manipulation_menu()
    except e.ReturnToMenu:
        # Exit diecase operations
        return True


def list_diecases():
    """Lists all matrix cases we have."""
    data = DB.get_all_diecases()
    results = {}
    UI.display('\n' +
               'No.'.ljust(4) +
               'Diecase ID'.ljust(25) +
               'Wedge'.ljust(12) +
               'Typeface' +
               '\n\n0 - start a new empty diecase\n')
    for index, diecase in enumerate(data, start=1):
        # Start each row with index
        row = [str(index).ljust(4)]
        # Collect the ciecase parameters: ID, typeface, wedge
        # Leave the diecase layout out
        row.append(diecase[0].ljust(25))
        # Swap the wedge and typeface designations (more place for long names)
        row.append(diecase[2].ljust(12))
        row.append(diecase[1])
        # Add number and ID to the result that will be returned
        results[index] = diecase[0]
        UI.display(''.join(row))
    UI.display('\n\n')
    # Now we can return the number - diecase ID pairs
    return results


def choose_diecase():
    """Lists diecases and lets the user choose one;
    returns the Diecase class object with all parameters set up."""
    prompt = 'Number of a diecase (0 for a new one, leave blank to exit): '
    while True:
        try:
            UI.display('Choose a matrix case:', end='\n\n')
            data = list_diecases()
            choice = UI.enter_data_or_blank(prompt, int)
            if choice == 0:
                return None
            elif not choice:
                e.return_to_menu()
            else:
                return DB.get_diecase(data[choice])
        except KeyError:
            UI.pause('Diecase number is incorrect!')
        except (e.NoMatchingData, e.DatabaseQueryError):
            UI.display('No diecases found in database')
            return None


def import_layout_file():
    """Reads a matrix case arrangement from a text or csv file.
    The format should be:
    "character";"style1,style2...";"column";"row";"unit_width"
    """
    # Give us a file or end here
    filename = UI.enter_input_filename()
    if not filename:
        return False
    # Initialize the records list
    all_records = []
    # This will store the processed combinations - and whenever a duplicate
    # is detected, the function will raise an exception
    with io.open(filename, 'r') as layout_file:
        input_data = csv.reader(layout_file, delimiter=';', quotechar='"')
        all_records = [record for record in input_data]
        displayed_lines = [' '.join(record) for record in all_records[:5]]
        # Preview file
        UI.display('File preview: displaying first 5 rows:\n')
        UI.display('\n'.join(displayed_lines), end='\n\n')
        # Ask if the first row is a header - if so, away with it
        if UI.confirm('Is the 1st row a table header? ', default=True):
            all_records.pop(0)
    if not UI.confirm('Proceed?', default=True):
        return False
    try:
        # Process the records
        processed_records = [process_record(record) for record in all_records]
        # Determine the diecase size based on row and column
        # Get columns and rows lists
        columns = {record[2] for record in processed_records}
        rows = sorted({record[3] for record in processed_records})
        # Check if 17 columns (15x17, 16x17), else 15 columns (old 15x15)
        big_layout = 'NI' in columns or 'NL' in columns or 16 in rows
        columns_list = big_layout and c.COLUMNS_17 or c.COLUMNS_15
        # We now have completed uploading a layout and making a list out of it
        layout = [record for row in rows for col in columns_list
                  for record in processed_records
                  if record[2] == col and record[3] == row]
        # Show the uploaded layout
        return layout
    except (KeyError, ValueError, IndexError):
        return False


def process_record(record):
    """Prepares the record found in file for adding to the layout"""
    # A record is a list with all diecase data:
    # [character, (style1, style2...), column, row, units]
    # Add a character - first item; if it's a space, don't change it
    try:
        # 5 fields in a record = unit value given
        # Unit value must be converted to int
        (char, styles, column, row, units) = record
        units = int(units.strip())
    except (ValueError, AttributeError):
        # 4 fields = unit value not given
        # (or unit value cannot be converted to int)
        (char, styles, column, row) = record
        units = 0
    if char != ' ':
        char = char.strip()
    styles = [style.strip() for style in styles.split(',')]
    row = int(row.strip())
    column = column.strip()
    # Pack it again
    return (char, styles, column, row, units)


def generate_empty_layout(rows=None, columns=None):
    """Makes a list of empty matrices, row for row, column for column"""
    prompt = ('Matrix case size: 1 for 15x15, 2 for 15x17, 3 for 16x17? '
              '(leave blank to cancel) : ')
    options = {'1': (15, 15), '2': (15, 17), '3': (16, 17), '': ()}
    if (rows, columns) not in options.values():
        choice = UI.simple_menu(prompt, options)
        if not choice:
            return
        (rows, columns) = choice
    # Generate column numbers
    columns_list = columns == 17 and c.COLUMNS_17 or c.COLUMNS_15
    # Generate row numbers: 1...15 or 1...16
    rows_list = [num + 1 for num in range(rows)]
    return [Matrix(coordinates=(column, row), units=0)
            for row in rows_list for column in columns_list]



# -*- coding: utf-8 -*-
"""Operations on ribbon and scheme objects: creating, editing and deleting. """
# File operations
import io
# CSV reader/writer
import csv
# Some functions raise custom exceptions
from . import exceptions as e
# Matrix data for diecases
from . import matrix_data
from . import wedge_data
# Constants for rpi2caster
from .constants import ASSIGNMENT_SYMBOLS
# User interface
from .global_settings import UI
# Database
from . import database
DB = database.Database()


class Ribbon(object):
    """Ribbon objects - no matter whether files or database entries.

    A ribbon has the following attributes:
    description, customer - strings
    diecase_id (diecase is selected automatically on casting, so user can e.g.
                view the layout or know which wedge to use)
    contents - series of Monotype codes

    Methods:
    choose_ribbon - choose ribbon automatically or manually,
        first try to get a ribbon with ribbon_id, and if that fails
        ask and select ribbon manually from database, and if that fails
        ask and load ribbon from file
    read_from_file - select a file, parse the metadata, set the attributes
    get_from_db - get data from database
    export_to_file - store the metadata and contents in text file
    store_in_db - store the metadata and contents in db
    set_[description, customer, diecase_id] - set parameters manually"""
    def __init__(self):
        self.ribbon_id = None
        self.description = 'No Description'
        self.customer = 'Default Customer'
        self.filename = 'New Ribbon'
        self.contents = []
        self.diecase = matrix_data.Diecase()
        self.wedge = None

    def __iter__(self):
        return iter(self.contents)

    def __next__(self):
        yield from self.contents

    def __repr__(self):
        return self.ribbon_id or ''

    def set_ribbon_id(self, ribbon_id=None):
        """Sets the ribbon ID"""
        prompt = 'Ribbon ID? (leave blank to exit) : '
        ribbon_id = (ribbon_id or UI.enter_data_or_blank(prompt) or
                     self.ribbon_id)
        # Ask if we are sure we want to update this
        # if self.ribbon_id was set earlier
        prompt = 'Are you sure to change the ribbon ID?'
        condition = not self.ribbon_id or UI.confirm(prompt, default=False)
        if condition:
            self.ribbon_id = ribbon_id
            return True

    def set_description(self, description=None):
        """Manually sets the ribbon's description"""
        prompt = 'Enter the title: '
        self.description = (description or
                            UI.enter_data_or_blank(prompt) or self.description)

    def set_customer(self, customer=None):
        """Manually sets the customer"""
        prompt = 'Enter the customer\'s name for this ribbon: '
        self.customer = (customer or
                         UI.enter_data_or_blank(prompt) or self.customer)

    def set_diecase(self, diecase_id=None):
        """Chooses the diecase for this ribbon"""
        self.diecase = matrix_data.SelectDiecase(diecase_id)

    def set_wedge(self, wedge_name=None):
        """Sets a wedge for the ribbon"""
        self.wedge = wedge_name

    @property
    def parameters(self):
        """Gets a list of parameters"""
        return [(self.filename, 'File name'),
                (self.ribbon_id, 'Ribbon ID'),
                (self.description, 'Description'),
                (self.customer, 'Customer'),
                (self.diecase.diecase_id, 'Matrix case ID'),
                (self.wedge, 'Wedge')]

    @property
    def wedge(self):
        """Gets the ribbon's wedge"""
        return self.__dict__.get('_wedge', wedge_data.Wedge())

    @wedge.setter
    def wedge(self, wedge_name):
        """Sets the wedge for this ribbon"""
        self.__dict__['_wedge'] = wedge_data.SelectWedge(wedge_name or '5-12E')

    def display_contents(self):
        """Displays the ribbon's contents, line after line"""
        UI.display('Ribbon contents preview:\n')
        contents_generator = (line for line in self.contents if line)
        try:
            while True:
                UI.display(contents_generator.__next__())
        except StopIteration:
            # End of generator
            UI.pause('Finished', UI.MSG_MENU)
        except (EOFError, KeyboardInterrupt):
            # Press ctrl-C to abort displaying long ribbons
            UI.pause('Aborted', UI.MSG_MENU)

    def get_from_db(self):
        """Gets the ribbon from database"""
        data = choose_ribbon_from_db()
        if data and UI.confirm('Override current data?', default=False):
            (self.ribbon_id, self.description, self.customer, dc_id,
             self.wedge, self.contents) = data
            self.diecase = dc_id and self.set_diecase(dc_id) or self.diecase
            return True

    def store_in_db(self):
        """Stores the ribbon in database"""
        UI.display_parameters({'Ribbon data': self.parameters})
        # Ask for confirmation
        try:
            DB.add_ribbon(self)
            UI.pause('Ribbon added successfully.')
            return True
        except (e.DatabaseQueryError, e.NoMatchingData):
            UI.pause('Cannot store ribbon in database!')
            return False

    def delete_from_db(self):
        """Deletes a ribbon from database."""
        if UI.confirm('Are you sure?', default=False):
            DB.delete_ribbon(self)
            UI.display('Ribbon deleted successfully.')

    def import_from_file(self, filename=None):
        """Reads a ribbon file, parses its contents, sets the ribbon attrs"""
        (self.ribbon_id, self.description, self.customer, dc_id,
         self.wedge, self.contents) = import_ribbon_from_file()
        # Set filename as attribute
        self.filename = filename
        self.diecase = dc_id and self.set_diecase(dc_id) or self.diecase

    def export_to_file(self, filename=None):
        """Exports the ribbon to a text file"""
        UI.display_parameters({'Ribbon data': self.parameters})
        # Choose file, write metadata, write contents
        filename = filename or UI.enter_output_filename()
        with io.open(filename, mode='w+') as ribbon_file:
            ribbon_file.write('description: ' + self.description)
            ribbon_file.write('customer: ' + self.customer)
            ribbon_file.write('diecase: ' + self.diecase.diecase_id)
            ribbon_file.write('wedge: ' + self.wedge)
            for line in self.contents:
                ribbon_file.write(line)


class SelectRibbon(Ribbon):
    """A class for ribbons chosen from database or file"""
    def __init__(self, ribbon_id=None, filename=None):
        super().__init__()
        try:
            ribbon_data = (ribbon_id and DB.get_ribbon(ribbon_id) or
                           filename and import_ribbon_from_file(filename) or
                           choose_ribbon_from_db() or
                           import_ribbon_from_file())
            (self.ribbon_id, self.description, self.customer, dc_id,
             self.wedge, self.contents) = ribbon_data
            # Update diecase if correct diecase_id found in ribbon
            self.diecase = dc_id and self.set_diecase(dc_id) or self.diecase
        # Otherwise use empty ribbon
        except (e.NoMatchingData, e.DatabaseQueryError):
            UI.display('Ribbon choice failed. Starting a new one.')


class FontScheme(object):
    """Font schemes store information about how many of each character is to be
    cast when casting typecases. Each language has its own."""
    def __init__(self):
        self.scheme_id = None
        self.description = None
        self.language = None
        self.layout = {}

    def __iter__(self):
        return iter(self.layout)

    def __next__(self):
        yield from self.layout

    def __repr__(self):
        return self.scheme_id or ''

    @property
    def parameters(self):
        """Gets a list of parameters"""
        return [(self.scheme_id, 'Font scheme ID'),
                (self.description, 'Description'),
                (self.language, 'Language')]

    def set_scheme_id(self, scheme_id=None):
        """Sets the scheme ID"""
        prompt = 'Scheme ID? (leave blank to exit) : '
        scheme_id = (scheme_id or UI.enter_data_or_blank(prompt) or
                     self.scheme_id)
        # Ask if we are sure we want to update this
        # if self.scheme_id was set earlier
        prompt = 'Are you sure to change the scheme ID?'
        condition = (not self.scheme_id or scheme_id != self.scheme_id and
                     UI.confirm(prompt, default=False))
        if condition:
            self.scheme_id = scheme_id
            return True

    def set_description(self, description=None):
        """Manually sets the scheme's description"""
        prompt = 'Enter the title: '
        self.description = (description or UI.enter_data_or_blank(prompt) or
                            self.description)

    def set_language(self, language=None):
        """Manually sets the language"""
        prompt = 'Enter the language for this font scheme: '
        self.language = (language or UI.enter_data_or_blank(prompt) or
                         self.language)

    def import_from_file(self, filename=None):
        """Reads a text file and sets contents."""
        self.layout = import_scheme_from_file(filename)

    def delete_from_db(self):
        """Deletes a scheme from database."""
        if UI.confirm('Are you sure?', default=False):
            try:
                DB.delete_scheme(self)
                UI.display('Font scheme definition deleted successfully.')
            except (e.NoMatchingData, e.DatabaseQueryError):
                UI.display('Cannot delete scheme from database!')

    def store_in_db(self):
        """Stores the scheme in database"""
        try:
            DB.add_scheme(self)
            UI.pause('Data added successfully.')
            return True
        except (e.DatabaseQueryError, e.NoMatchingData):
            UI.pause('Cannot store scheme in database!')


class SelectFontScheme(FontScheme):
    """A class for font schemes selected from database or file"""
    def __init__(self, scheme_id=None, filename=None):
        super().__init__()
        try:
            scheme_data = (scheme_id and DB.get_scheme(scheme_id) or
                           choose_scheme_from_db())
            (self.scheme_id, self.description, self.language,
             self.layout) = scheme_data
        except (TypeError, e.NoMatchingData, e.DatabaseQueryError):
            UI.display('Scheme choice failed. Starting a new one.')
            self.layout = import_scheme_from_file(filename)


def list_ribbons():
    """Lists all ribbons in the database."""
    data = DB.get_all_ribbons()
    results = {}
    UI.display('\n' +
               'No.'.ljust(5) +
               'Ribbon name'.ljust(30) +
               'Description'.ljust(30) +
               'Customer'.ljust(30) +
               'Diecase'.ljust(30) +
               'Wedge'.ljust(30) +
               '\n\n0 - start a new empty ribbon\n')
    for index, ribbon in enumerate(data, start=1):
        # Collect ribbon parameters
        row = [str(index).ljust(5)]
        # Add ribbon name, description, diecase ID
        row.extend([field.ljust(30) for field in ribbon[:-1]])
        # Display it all
        UI.display(''.join(row))
        # Add number and ID to the result that will be returned
        results[index] = ribbon[0]
    UI.display('\n\n')
    # Now we can return the number - ribbon ID pairs
    return results


def list_schemes():
    """Lists all font schemes in the database."""
    data = DB.get_all_schemes()
    results = {}
    UI.display('\n' +
               'No.'.ljust(5) +
               'Scheme ID'.ljust(20) +
               'Description'.ljust(20) +
               'Language'.ljust(20) +
               '\n\n0 - start a new empty ribbon\n')
    for index, scheme in enumerate(data, start=1):
        # Collect scheme parameters
        row = [str(index).ljust(5)]
        row.extend([field.ljust(20) for field in scheme[:-1]])
        UI.display(''.join(row))
        # Add number and ID to the result that will be returned
        results[index] = scheme[0]
    UI.display('\n\n')
    # Now we can return the number - scheme ID pairs
    return results


def import_ribbon_from_file(filename=None):
    """Imports ribbon from file, parses parameters, returns a list of them"""
    def get_value(line, symbol):
        """Helper function - strips whitespace and symbols"""
        # Split the line on an assignment symbol, get the second part,
        # strip any whitespace or multipled symbols
        return line.split(symbol, 1)[-1].strip(symbol).strip()

    UI.display('Loading the ribbon from file...')
    try:
        filename = filename or UI.enter_input_filename()
        with io.open(filename, mode='r') as ribbon_file:
            ribbon = [line.strip() for line in ribbon_file if line.strip()]
    except (FileNotFoundError, IOError):
        UI.pause('Cannot open ribbon file %s' % filename)
        return False
    # What to look for
    keywords = ['diecase', 'description', 'desc', 'diecase_id', 'customer',
                'wedge', 'stopbar']
    # Metadata (anything found), contents (the rest)
    metadata = {}
    contents = []
    # Look for parameters line per line, get parameter value
    # If parameters exhausted, append the line to contents
    for line in ribbon:
        for keyword in keywords:
            if line.startswith(keyword):
                for sym in ASSIGNMENT_SYMBOLS:
                    if sym in line:
                        # Data found
                        metadata[keyword] = get_value(line, sym)
                        break
                break
        else:
            contents.append(line)
    # Metadata parsing
    ribbon_id = None
    description = metadata.get('description', '') or metadata.get('desc', '')
    customer = metadata.get('customer', '')
    diecase_id = metadata.get('diecase', '') or metadata.get('diecase_id', '')
    wedge = metadata.get('wedge', '') or metadata.get('stopbar', '')
    # Add the whole contents as the attribute
    return (ribbon_id, description, customer, diecase_id, wedge, contents)


def import_scheme_from_file(filename=None):
    """Imports scheme text from file"""
    filename = filename or UI.enter_input_filename()
    # Initialize the contents
    with io.open(filename, mode='r') as scheme_file:
        input_data = csv.reader(scheme_file, delimiter=';', quotechar='"')
        all_records = [record for record in input_data]
    displayed_lines = [' '.join(record) for record in all_records[:5]]
    # Preview file
    UI.display('File preview: displaying first 5 rows:\n')
    UI.display('\n'.join(displayed_lines), end='\n\n')
    # Ask if the first row is a header - if so, away with it
    if UI.confirm('Is the 1st row a table header? ', default=True):
        all_records.pop(0)
    try:
        return {char: int(qty) for char, qty in all_records}
    except (KeyError, ValueError, IndexError):
        return {}


def choose_ribbon_from_db():
    """Chooses ribbon data from database"""
    prompt = 'Number of a ribbon? (0 for a new one, leave blank to exit): '
    while True:
        try:
            data = list_ribbons()
            choice = UI.enter_data_or_blank(prompt, int)
            if choice == 0:
                return None
            elif not choice:
                e.return_to_menu()
            else:
                return DB.get_ribbon(data[choice])
        except KeyError:
            UI.pause('Ribbon number is incorrect!')
        except (e.DatabaseQueryError, e.NoMatchingData):
            UI.display('No ribbons found in database')
            return None


def choose_scheme_from_db():
    """Chooses scheme data from database"""
    prompt = 'Number of a scheme? (0 for a new one, leave blank to exit): '
    while True:
        try:
            data = list_schemes()
            choice = UI.enter_data_or_blank(prompt, int)
            if choice == 0:
                return None
            elif not choice:
                e.return_to_menu()
            else:
                return DB.get_scheme(data[choice])
        except KeyError:
            UI.pause('Font scheme number is incorrect!')
        except (e.DatabaseQueryError, e.NoMatchingData):
            UI.display('No font schemes found in database')
            return None



# -*- coding: utf-8 -*-
"""
casting:

A module for everything related to working on a Monotype composition caster:
-casting composed type,
-punching paper tape (ribbon) for casters without interfaces,
-casting l lines of m sorts from matrix with x, y coordinates,
-composing and casting a line of text
-testing all valves, lines and pinblock,
-calibrating the space transfer wedge,
-heating the mould up,
-sending any codes/combinations to the caster and keeping them on.
"""

# IMPORTS:
from collections import deque
# Signals parsing methods for rpi2caster
from . import parsing as p
# Custom exceptions
from . import exceptions as e
# Constants shared between modules
from . import constants as c
# Typesetting functions module
# from . import typesetting_funcs
# Caster backend
from . import monotype
# Casting stats
from .casting_stats import Stats
# Globally selected UI
from .global_settings import UI
# Matrix, wedge and typesetting data models
from . import typesetting_data
from . import matrix_data
from . import wedge_data
# Constants for control sequences
GALLEY_TRIP = 'NKJS 0005 0075'
PUMP_OFF = 'NJS 0005'
PUMP_ON = 'NKS 0075'
END_CASTING = [GALLEY_TRIP, PUMP_OFF, PUMP_OFF]


def choose_sensor_and_driver(casting_routine):
    """Checks current modes (simulation, perforation, testing)"""
    def wrapper(self, *args, **kwargs):
        """Wrapper function"""
        UI.debug_pause('About to %s...' %
                       (self.caster.mode.casting and 'cast composition' or
                        self.caster.mode.testing and 'test the outputs' or
                        self.caster.mode.calibration and
                        'calibrate the machine' or
                        self.caster.mode.punching and 'punch the ribbon'))
        # Instantiate and enter context
        with self.caster.mode.sensor() as self.caster.sensor:
            with self.caster.mode.output() as self.caster.output:
                with self.caster:
                    return casting_routine(self, *args, **kwargs)
    return wrapper


def cast_or_punch_result(ribbon_source):
    """Get the ribbon from decorated routine and cast it"""
    def wrapper(self, *args, **kwargs):
        """Wrapper function"""
        try:
            self.cast_queue(ribbon_source(self, *args, **kwargs))
        except e.CastingAborted:
            pass
        self.caster.mode.diagnostics = False
        self.caster.mode.hmn, self.caster.mode.unitshift = False, False
    return wrapper


def prepare_job(ribbon_casting_workflow):
    """Prepares the job for casting"""

    def wrapper(self, ribbon):
        """Wrapper function"""
        # Stop here if no ribbon
        if not ribbon:
            return
        # Mode aliases
        punching = self.caster.mode.punching
        casting = self.caster.mode.casting
        diagnostics = self.caster.mode.diagnostics
        line_of_quads = ['O15'] * 20 + ['NKJS 0005 0075']
        # Rewind the ribbon if 0005 is found before 0005+0075
        if not diagnostics and not punching and p.stop_comes_first(ribbon):
            ribbon = [x for x in reversed(ribbon)]
        # New stats for the resulting ribbon
        self.stats.ribbon = ribbon
        UI.display_parameters({'Ribbon info': self.stats.ribbon_parameters})
        # Always 1 run for calibrating and punching
        prompt = 'How many times do you want to cast it? (default: 1) : '
        self.stats.runs = abs(not diagnostics and not punching and
                              UI.enter_data_or_blank(prompt, int) or 1)
        # Line skipping - ask user if they want to skip any initial line(s)
        prompt = 'How many initial lines do you want to skip?  (default: 0): '
        l_skipped = abs(not diagnostics and not punching and
                        UI.enter_data_or_blank(prompt, int) or 0)
        # Leave at least one line in the ribbon to cast
        l_skipped = min(l_skipped, self.stats.get_ribbon_lines() - 1)
        UI.display_parameters({'Session info': self.stats.session_parameters})
        # For each casting run repeat
        while self.stats.get_runs_left():
            queue = deque(ribbon)
            quad_lines = 0
            if (not punching and not diagnostics and
                    self.stats.get_current_run() < 2 and
                    UI.confirm('Cast 2 extra lines to heat up the mould?',
                               default=False)):
                l_skipped -= 2
                # When less than 2 lines can be skipped, we'll cast quads
                quad_lines = max(-l_skipped, 0)
            # Apply constraints: 0 <= lines_skipped < lines in ribbon
            l_skipped = max(0, l_skipped)
            UI.display(l_skipped and ('Skipping %s lines' % l_skipped) or '')
            prompt = 'Casting %s line(s) of 20 quads' % quad_lines
            UI.display(quad_lines and prompt or '')
            # Make sure we take off lines from the beginning of the
            # casting job - i.e. last lines of text; count the lines
            code = ''
            # Add one more line to skip - because ribbon starts with 0075-0005
            # Guard against negative infinite loop if l_skipped == -2
            while l_skipped + 1 > 0 and not diagnostics:
                code = queue.popleft()
                l_skipped -= 1 * p.check_newline(code)
            # Add it back, then add those quads
            queue.appendleft(code)
            queue.extendleft(quad_lines * line_of_quads)
            # The ribbon is ready for casting / punching
            self.stats.queue = queue
            exit_prompt = '[Y] to start next run or [N] to exit?'
            if ribbon_casting_workflow(self, queue):
                # Casting successful - ready to cast next run - ask to repeat
                # after the last run is completed (because user may want to
                # cast / punch once more?)
                if (self.stats.all_done() and
                        UI.confirm('One more run?', default=False)):
                    self.stats.add_one_more_run()
            elif casting and UI.confirm('Retry this run?', default=True):
                # Casting aborted - ask if user wants to repeat
                self.stats.undo_last_run()
                self.stats.add_one_more_run()
                lines_ok = self.stats.get_lines_done()
                prompt = 'Skip %s lines successfully cast?' % lines_ok
                if lines_ok > 0 and UI.confirm(prompt, default=True):
                    l_skipped = lines_ok
                # Start this run again
            elif (not self.stats.all_done() and
                    UI.confirm(exit_prompt, default=True)):
                # There are some more runs to do - go on?
                self.stats.undo_last_run()
            else:
                return
    return wrapper


class Casting(object):
    """Casting:

    Methods related to operating the composition caster.
    Requires configured caster.

    All methods related to operating a composition caster are here:
    -casting composition and sorts, punching composition,
    -calibrating the caster,
    -testing the interface,
    -sending an arbitrary combination of signals,
    -casting spaces to heat up the mould."""

    def __init__(self, ribbon_file='', diecase='', wedge=''):
        # Caster for this job
        self.caster = monotype.MonotypeCaster()
        self.stats = Stats(self)
        self.ribbon = (ribbon_file and
                       typesetting_data.SelectRibbon(filename=ribbon_file) or
                       typesetting_data.Ribbon())
        self.diecase = (diecase and
                        matrix_data.SelectDiecase(diecase) or self.diecase)
        self.wedge = wedge and wedge_data.SelectWedge(wedge) or self.wedge

    @choose_sensor_and_driver
    @prepare_job
    def cast_queue(self, casting_queue):
        """Casts the sequence of codes in ribbon or self.ribbon.contents,
        displaying the statistics (depending on context:
        casting, punching or testing)
        """
        generator = (p.parse_record(record) for record in casting_queue)
        if not self.caster.mode.testing:
            self.caster.sensor.check_if_machine_is_working()
        while True:
            try:
                signals, comment = next(generator)
                if comment and not signals:
                    UI.display('\n\n' + comment + '\n' + '-' * len(comment))
                    continue
                # Check if HMN or unit-shift must be applied
                self.stats.signals = signals
                UI.display_parameters({comment: self.stats.code_parameters})
                # Let the caster do the job
                self.caster.process(signals)
            except StopIteration:
                self.stats.end_run()
                return True
            except (e.MachineStopped, KeyboardInterrupt, EOFError):
                # Allow resume in punching mode
                if (self.caster.mode.punching and not
                        self.caster.mode.diagnostics and
                        UI.confirm('Continue?', default=True)):
                    self.caster.process(signals)
                else:
                    self.stats.end_run()
                    return False

    @cast_or_punch_result
    def _test_front_pinblock(self):
        """Sends signals 1...14, one by one"""
        UI.pause('Testing the front pinblock - signals 1 towards 14.')
        self.caster.mode.testing = True
        return [str(n) for n in range(1, 15)]

    @cast_or_punch_result
    def _test_rear_pinblock(self):
        """Sends NI, NL, A...N"""
        UI.pause('This will test the front pinblock - signals NI, NL, A...N. ')
        self.caster.mode.testing = True
        return [x for x in c.COLUMNS_17]

    @cast_or_punch_result
    def _test_all(self):
        """Tests all valves and composition caster's inputs in original
        Monotype order: NMLKJIHGFSED 0075 CBA 123456789 10 11 12 13 14 0005.
        """
        UI.pause('This will test all the air lines in the same order '
                 'as the holes on the paper tower: \n%s\n'
                 'MAKE SURE THE PUMP IS DISENGAGED.' % ' '.join(c.SIGNALS))
        self.caster.mode.testing = True
        return [x for x in c.SIGNALS]

    @cast_or_punch_result
    def _test_justification(self):
        """Tests the 0075-S-0005"""
        UI.pause('This will test the justification pin block (0075, S, 0005).')
        self.caster.mode.testing = True
        return ['0075', 'S', '0005']

    @choose_sensor_and_driver
    def _test_any_code(self):
        """Tests a user-specified combination of signals"""
        self.caster.mode.testing = True
        while True:
            UI.display('Enter the signals to send to the caster, '
                       'or leave empty to return to menu: ')
            prompt = 'Signals? (leave blank to exit): '
            string = UI.enter_data_or_blank(prompt)
            signals = p.parse_signals(string)
            if signals:
                self.caster.output.valves_off()
                UI.display('Activating ' + ' '.join(signals))
                self.caster.output.valves_on(signals)
            else:
                break
        self.caster.mode.testing = False

    @cast_or_punch_result
    def cast_composition(self):
        """Casts or punches the ribbon contents if there are any"""
        if not self.ribbon.contents:
            e.return_to_menu()
        return self.ribbon.contents

    @cast_or_punch_result
    def cast_sorts(self, source=()):
        """Sorts casting routine, based on the position in diecase.
        Ask user about the diecase row & column, as well as number of sorts.
        """
        queue = []
        casting_queue = [x for x in source] or [(None, None, 0)]
        for (char, style, qty) in casting_queue:
            # Make an infinite iterator when we call this function without
            # the list of character-style pairs
            prompt = ('Character: %s %s\n'
                      'Y: add to casting queue, N: skip it?' % (style, char))
            if not char:
                # Make it an infinite queue if no input specified
                # (cast any number of sorts we want)
                casting_queue.append((char, style, 0))
            elif not UI.confirm(prompt, default=True):
                continue
            if self.diecase:
                prompt = ('Character? (leave blank to end '
                          'specifying characters and start casting): ')
                char = char or UI.enter_data_or_blank(prompt)
                if not char:
                    break
                style = style or UI.choose_one_style()
                matrix = self.diecase.get_matrix(char, style)
            else:
                matrix = self.diecase.get_matrix()
            prompt = 'Unit correction (-2...+10, default 0) ?: '
            correction = 20
            while not -2 <= correction <= 10:
                try:
                    correction = abs(int(UI.enter_data_or_blank(prompt) or 0))
                except (ValueError, TypeError):
                    correction = 20
            diff = matrix.units + correction - matrix.row_units(self.wedge)
            # Calculate wedges
            wedge_positions = matrix.wedge_positions(correction, self.wedge)
            # Ask for number of sorts and lines, no negative numbers here
            prompt = '\nHow many sorts per line? (default: 10): '
            qty = qty or abs(UI.enter_data_or_blank(prompt, int) or 10)
            prompt = '\nHow many lines? (default: 1): '
            lines = abs(UI.enter_data_or_blank(prompt, int) or 1)
            UI.display('\n')
            # Warn if we want to cast too many sorts from a single matrix
            warning = ('Warning: you want to cast a single character more than'
                       ' 10 times. This may lead to matrix overheating!\n')
            if qty > 10:
                UI.display(warning)
            UI.display('\n')
            line_codes = [GALLEY_TRIP + str(wedge_positions['0005']),
                          PUMP_ON + str(wedge_positions['0075'])] + ['O15'] * 2
            # 2 quads in the beginning, then spaces, then 2 quads in the end
            line_codes.extend([matrix.code + (diff and 'S' or '') +
                               '// ' + matrix.char] * qty)
            line_codes.extend(['O15'] * 2)
            # Ask for confirmation
            queue.extend(line_codes * lines)
            prompt = 'Add next character? If not, cast the queue...'
            if not (self.diecase or
                    UI.confirm(prompt, default=True)):
                break
        # No sequences? No casting.
        if queue:
            UI.display('\nEach line will have two em-quads at the start '
                       'and at the end, to support the type.\n')
            return queue + END_CASTING
        else:
            e.return_to_menu()

    @cast_or_punch_result
    def cast_spaces(self):
        """Spaces casting routine, based on the position in diecase.
        Ask user about the space width and measurement unit.
        """
        queue = []
        line_length = (UI.enter_line_length() * 0.1667 / self.wedge.pica *
                       self.wedge.set_width / 12)
        # We need 72 additional units for two quads before and after each line
        line_length -= 72
        while True:
            matrix = self.diecase.get_matrix(' ')
            prompt = ('Space width? [6] = 1/6em, [4] = 1/4em, [3] = 1/3em, '
                      '[2] = 1/2em, [1] = 1em, [C] for custom width: ')
            # Width in points
            width = UI.simple_menu(prompt, {'6': 1/6, '4': 1/4, '3': 1/3,
                                            '2': 1/2, '1': 1, 'C': 0}) * 12.0
            # Ask about custom value, then specify units
            while not 2 <= width <= 20:
                prompt = 'Custom width in points (decimal, 2...20) ? : '
                width = UI.enter_data_or_blank(prompt, float)
            prompt = '\nHow many lines? (default: 1): '
            lines = abs(UI.enter_data_or_blank(prompt, int) or 1)
            prompt = ('Casting %s lines of %s-point spaces from %s. OK?'
                      % (lines, width, matrix))
            # Save for later
            comment = '%s-point space' % width
            # Repeat
            if not UI.confirm(prompt, default=True):
                continue
            # Calculate unit width of said number of points
            # We do it this way:
            # units = picas * set_width/12 * 18
            # a pica is 12 points, so:
            # units = width (points) * set_width/12 * 1 / 12 * 18
            # 18 / (12*12) = 0.125, hence division by 8
            width = round(width * 0.1667 / self.wedge.pica *
                          self.wedge.set_width / 8, 2)
            # How many spaces will fit in a line? Calculate it...
            qty = int(line_length // width)
            # Check if the corrections are needed at all
            correction = width - matrix.units
            wedge_positions = matrix.wedge_positions(correction, self.wedge)
            # Any corrections needed?
            diff = matrix.units + correction - matrix.row_units(self.wedge)
            # Add 'S' if there is width difference
            signals = matrix.code + (diff and 'S' or '') + '// ' + comment
            line_codes = [GALLEY_TRIP + str(wedge_positions['0005']),
                          PUMP_ON + str(wedge_positions['0075'])]
            # 2 quads in the beginning, then spaces, then 2 quads in the end
            line_codes.extend(['O15'] * 2 + [signals] * qty + ['O15'] * 2)
            # Ask for confirmation
            queue.extend(line_codes * lines)
            prompt = 'More spaces? Otherwise, start casting'
            if queue and not UI.confirm(prompt, default=True):
                break
        if queue:
            return queue + END_CASTING
        else:
            e.return_to_menu()

    def cast_typecases(self):
        """Casting typecases according to supplied font scheme."""
        enter = UI.enter_data_or_blank
        scheme = typesetting_data.SelectFontScheme().layout
        scale_prompt = ('Scale for %s in %%, relative to 100 "a" characters:\n'
                        '(default: 100%%) ?: ')
        UI.display('Styles to cast?')
        styles = UI.choose_styles()
        # Now generate and display the casting queue
        style_scales = {style: (enter(scale_prompt % style, float) / 100) or 1
                        for style in styles}
        # Get the char number from scheme, style scale from scales
        queue = [(char, [style], int(scheme[char] * style_scales[style]))
                 for style in styles for char in sorted(scheme)]
        queue = []
        for style in styles:
            UI.display_header(style)
            for char in sorted(scheme):
                qty = int(scheme[char] * style_scales[style])
                UI.display('%s : %s' % (char, qty))
                queue.append((char, style, qty))
        self.cast_sorts(queue)

    @cast_or_punch_result
    def adhoc_typesetting(self):
        """Allows us to use caster for casting single lines.
        This means that the user enters a text to be cast,
        gives the line length, chooses alignment and diecase.
        Then, the program composes the line, justifies it, translates it
        to Monotype code combinations.

        This allows for quick typesetting of short texts, like names etc.
        """
        # Initialize the typesetter for a chosen diecase
        pass

    @cast_or_punch_result
    def _calibrate_wedges(self):
        """Allows to calibrate the justification wedges so that when you're
        casting a 9-unit character with the S-needle at 0075:3 and 0005:8
        (neutral position), the    width is the same.

        It works like this:
        1. 0075 - turn the pump on,
        2. cast 10 spaces from the specified matrix (default: G9),
        3. put the line to the galley & set 0005 to 8, 0075 to 3, pump on,
        4. cast 10 spaces with the S-needle from the same matrix,
        5. put the line to the galley, then 0005 to turn the pump off.
        """
        UI.display('Transfer wedge calibration:\n\n'
                   'This function will cast two lines of 5 spaces: '
                   'first: G5, second: GS5 with wedges at 3/8. \n'
                   'Adjust the 52D space transfer wedge '
                   'until the lengths are the same.')
        if not UI.confirm('\nProceed?', default=True):
            return None
        signals = 'G5'
        queue = [GALLEY_TRIP] + [signals] * 7
        queue.extend([GALLEY_TRIP + '8', PUMP_ON + '3'] + [signals + 'S'] * 7)
        queue.extend(END_CASTING)
        self.caster.mode.calibration = True
        return queue

    @cast_or_punch_result
    def _calibrate_mould(self):
        """Calculates the width, displays it and casts some 9-unit characters.
        Then, the user measures the width and adjusts the mould opening width.
        """
        UI.display('Mould blade opening calibration:\n'
                   'Cast G5 (9-units wide on S5 wedge), then measure '
                   'the width. Adjust if needed.')
        UI.display_parameters({'Wedge data': self.wedge.parameters})
        UI.display('\n9 units (1en) is %s" wide'
                   % round(self.wedge.em_width / 2, 4))
        UI.display('18 units (1em) is %s" wide\n'
                   % round(self.wedge.em_width, 4))
        if not UI.confirm('\nProceed?', default=True):
            return None
        signals = 'G5'
        self.caster.mode.calibration = True
        return [GALLEY_TRIP] + [signals] * 7 + END_CASTING

    def _calibrate_diecase(self):
        """Casts the "en dash" characters for calibrating the character X-Y
        relative to type body."""
        UI.pause('X-Y character calibration:\n'
                 'Cast some en-dashes and/or lowercase "n" letters, '
                 'then check the position of the character relative to the '
                 'type body.\nAdjust if needed.')
        self.caster.mode.calibration = True
        self.cast_sorts([('–', 'roman', 7), ('n', 'roman', 7),
                         ('h', 'roman', 7)])

    @choose_sensor_and_driver
    def _calibrate_draw_rods(self):
        """Keeps the diecase at G8 so that the operator can adjust
        the diecase draw rods until the diecase stops moving sideways
        when the centering pin is descending."""
        UI.display('Draw rods calibration:\n'
                   'The diecase will be moved to the central position (G-8), '
                   'turn on the machine\nand adjust the diecase draw rods '
                   'until the diecase stops moving sideways as the\n'
                   'centering pin is descending into the hole in the matrix.')
        if not UI.confirm('\nProceed?', default=True):
            return None
        self.caster.output.valves_on(['G', '8'])
        UI.pause('Sending G8, waiting for you to stop...')

    def diagnostics_submenu(self):
        """Settings and alignment menu for servicing the caster"""
        def menu_options():
            """Build a list of options, adding an option if condition is met"""
            caster = not self.caster.mode.punching
            opts = [(e.menu_level_up, 'Back',
                     'Returns to main menu', True),
                    (self._test_all, 'Test outputs',
                     'Tests all the air outputs one by one', True),
                    (self._test_front_pinblock, 'Test the front pinblock',
                     'Tests the pins 1...14', caster),
                    (self._test_rear_pinblock, 'Test the rear pinblock',
                     'Tests the pins NI, NL, A...N one by one', caster),
                    (self._test_justification, 'Test the 0075-S-0005 pinblock',
                     'Tests the pins for justification wedges', caster),
                    (self._test_any_code, 'Send specified signals',
                     'Sends the specified signal combination', True),
                    (self._calibrate_wedges, 'Calibrate the 52D wedge',
                     'Calibrate the space transfer wedge for correct width',
                     caster),
                    (self._calibrate_mould, 'Calibrate mould opening',
                     'Casts 9-unit characters to adjust the type width',
                     caster),
                    (self._calibrate_draw_rods, 'Calibrate diecase draw rods',
                     'Keep the matrix case at G8 and adjust the draw rods',
                     caster),
                    (self._calibrate_diecase, 'Calibrate matrix X-Y',
                     'Calibrate the character-to-body positioning', caster)]
            return [(function, description, long_description) for
                    (function, description, long_description, condition)
                    in opts if condition]

        header = ('Diagnostics and machine calibration menu:\n\n')
        # Keep displaying the menu and go back here after any method ends
        while True:
            try:
                # Catch "return to menu" and "exit program" exceptions here
                UI.menu(menu_options(), header=header)()
            except e.ReturnToMenu:
                # Stay in the menu
                pass

    def _main_menu_options(self):
        """Build a list of options, adding an option if condition is met"""
        # Options are described with tuples: (function, description, condition)
        caster = not self.caster.mode.punching
        diecase = self.diecase.diecase_id
        ribbon = self.ribbon.contents
        opts = [(e.exit_program, 'Exit', 'Exits the program', True),
                (self.cast_composition, 'Cast or punch composition',
                 'Casts type or punch a ribbon', ribbon),
                (self._choose_ribbon, 'Select ribbon',
                 'Selects a ribbon from database or file', True),
                (self._choose_diecase, 'Select diecase',
                 'Selects a matrix case from database', True),
                (self._choose_wedge, 'Select wedge',
                 'Selects a wedge from database', True),
                (self.ribbon.display_contents, 'View codes',
                 'Displays all sequences in a ribbon', ribbon),
                (self.diecase.show_layout, 'Show diecase layout',
                 'Views the matrix case layout', diecase),
                # (self.adhoc_typesetting, 'Ad-hoc typesetting',
                # 'Compose and cast a line of text', self.diecase),
                (self.cast_sorts, 'Cast sorts',
                 'Cast from matrix with given coordinates', caster),
                (self.cast_spaces, 'Cast spaces or quads',
                 'Casts spaces or quads of a specified width', caster),
                (self.cast_typecases, 'Cast typecases',
                 'Casts a typecase based on a selected font scheme',
                 caster and diecase),
                (self._display_details, 'Show detailed info...',
                 'Displays caster, ribbon, diecase and wedge details', True),
                (matrix_data.diecase_operations, 'Matrix manipulation...',
                 'Work on matrix cases', True),
                (self.diagnostics_submenu, 'Service...',
                 'Interface and machine diagnostic functions', True)]
        # Built a list of menu options conditionally
        return [(function, description, long_description)
                for (function, description, long_description, condition)
                in opts if condition]

    def main_menu(self):
        """Main menu for the type casting utility."""
        header = ('rpi2caster - CAT (Computer-Aided Typecasting) '
                  'for Monotype Composition or Type and Rule casters.\n\n'
                  'This program reads a ribbon (from file or database) '
                  'and casts the type on a composition caster.'
                  '\n\nCasting / Punching Menu:')
        # Keep displaying the menu and go back here after any method ends
        while True:
            # Catch any known exceptions here
            try:
                UI.menu(self._main_menu_options(), header=header, footer='')()
            except (e.ReturnToMenu, e.MenuLevelUp):
                # Will skip to the end of the loop, and start all over
                pass

    @property
    def ribbon(self):
        """Ribbon for the casting session"""
        return self.__dict__.get('_ribbon', typesetting_data.Ribbon())

    @ribbon.setter
    def ribbon(self, ribbon):
        """Ribbon setter"""
        self.__dict__['_ribbon'] = ribbon or typesetting_data.Ribbon()
        self.diecase = ribbon.diecase
        self.wedge = ribbon.wedge

    @property
    def diecase(self):
        """Diecase for the casting session"""
        return self.__dict__.get('_diecase', matrix_data.Diecase())

    @diecase.setter
    def diecase(self, diecase):
        """Diecase setter"""
        self.__dict__['_diecase'] = diecase or matrix_data.Diecase()
        self.wedge = diecase.wedge

    @property
    def wedge(self):
        """Wedge for the casting session"""
        return self.__dict__.get('_wedge', wedge_data.Wedge())

    @wedge.setter
    def wedge(self, wedge):
        """Wedge setter"""
        self.__dict__['_wedge'] = wedge or wedge_data.SelectWedge()

    def _choose_ribbon(self):
        """Chooses a ribbon from database or file"""
        self.ribbon = typesetting_data.SelectRibbon()

    def _choose_diecase(self):
        """Chooses a diecase from database"""
        self.diecase = matrix_data.SelectDiecase()

    def _choose_wedge(self):
        """Chooses a wedge from registered ones"""
        self.wedge = wedge_data.SelectWedge()

    def _display_details(self):
        """Collect ribbon, diecase and wedge data here"""
        UI.display_parameters({'Caster data': self.caster.parameters,
                               'Ribbon data': self.ribbon.parameters,
                               'Matrix case data': self.diecase.parameters,
                               'Wedge data': self.wedge.parameters})
        UI.pause()


def single_justification(pos_0075, pos_0005):
    """Returns a single justification sequence"""
    return [PUMP_OFF + str(pos_0005), PUMP_ON + str(pos_0075)]


def double_justification(pos_0075, pos_0005):
    """Returns a galley trip / double justification sequence"""
    return [GALLEY_TRIP + str(pos_0005), PUMP_ON + str(pos_0075)]



# -*- coding: utf-8 -*-
"""Constants shared between modules in the rpi2caster package"""

# Build a sequence of Monotype signals as they appear on the paper tower
SIGNALS = ([x for x in 'NMLKJIHGFSED'] + ['0075'] + [x for x in 'CBA'] +
           [str(x) for x in range(1, 15)] + ['0005', 'O15'])
# Signals corresponding to MCP0 A0...A7, B0...B7, MCP1 A0...A7, B0...B7
ALNUM_ARR = ('1,2,3,4,5,6,7,8,9,10,11,12,13,14,0005,'
             '0075,A,B,C,D,E,F,G,H,I,J,K,L,M,N,S,O15')

# Column numbers in 15x17 diecase
COLUMNS_15 = [x for x in 'ABCDEFGHIJKLMNO']
COLUMNS_17 = ['NI', 'NL'] + COLUMNS_15

# Styles
STYLES = {'r': 'roman', 'b': 'bold', 'i': 'italic',
          's': 'smallcaps', 'l': 'subscript', 'u': 'superscript'}

# For parsing lines in ribbon files
COMMENT_SYMBOLS = ['**', '*', '//', '##', '#']
ASSIGNMENT_SYMBOLS = ['=', ':', ' ']

# Useful for config files (all aliases must be lowercase)
TRUE_ALIASES = ['true', 'on', 'yes']
FALSE_ALIASES = ['false', 'off', 'no']

# Config file locations
DEFAULT_CONFIG_PATHS = ['/etc/rpi2caster.conf', 'data/rpi2caster.conf']
# Database file locations
DEFAULT_DATABASE_PATHS = ['/var/local/rpi2caster/monotype.db',
                          '/var/local/rpi2caster/rpi2caster.db',
                          'data/rpi2caster.db']

# Control codes
PUMP_STOP = ['N', 'J', '0005', 'S']
PUMP_START = ['N', 'K', '0075', 'S']
GALLEY_TRIP = ['N', 'K', 'J', '0005', '0075', 'S']
